"""
pyart.map.grid_mapper
=====================

Utilities for mapping radar objects to Cartesian grids.

.. autosummary::
    :toctree: generated/

    grid_from_radars
    map_to_grid
    example_roi_func_constant
    example_roi_func_dist
    _unify_times_for_radars
    _load_nn_field_data
    _gen_roi_func_constant
    _gen_roi_func_dist
    _gen_roi_func_dist_beam

.. autosummary::
    :toctree: generated/
    :template: dev_template.rst

    NNLocator

"""

import warnings

import netCDF4
import numpy as np
import scipy.spatial

from ..config import get_fillvalue, get_metadata
from ..core.transforms import geographic_to_cartesian
from ..core.grid import Grid
from ..core.radar import Radar
from ..filters import GateFilter, moment_based_gate_filter
from ..io.common import make_time_unit_str
from ._load_nn_field_data import _load_nn_field_data
from .ckdtree import cKDTree
from .gates_to_grid import map_gates_to_grid


def grid_from_radars(radars, grid_shape, grid_limits,
                     gridding_algo='map_gates_to_grid', **kwargs):
    """
    Map one or more radars to a Cartesian grid returning a Grid object.

    Additional arguments are passed to :py:func:`map_to_grid` or
    :py:func:`map_gates_to_grid`.

    Parameters
    ----------
    radars : Radar or tuple of Radar objects.
        Radar objects which will be mapped to the Cartesian grid.
    grid_shape : 3-tuple of floats
        Number of points in the grid (z, y, x).
    grid_limits : 3-tuple of 2-tuples
        Minimum and maximum grid location (inclusive) in meters for the
        z, y, x coordinates.
    gridding_algo : 'map_to_grid' or 'map_gates_to_grid'
        Algorithm to use for gridding. 'map_to_grid' finds all gates within
        a radius of influence for each grid point, 'map_gates_to_grid' maps
        each radar gate onto the grid using a radius of influence and is
        typically significantly faster.

    Returns
    -------
    grid : Grid
        A :py:class:`pyart.io.Grid` object containing the gridded radar
        data.

    See Also
    --------
    map_to_grid : Map to grid and return a dictionary of radar fields.
    map_gates_to_grid : Map each gate onto a grid returning a dictionary of
                        radar fields.

    References
    ----------
    Barnes S., 1964: A Technique for Maximizing Details in Numerical Weather
    Map Analysis. Journal of Applied Meteorology and Climatology, 3(4),
    396-409.

    Cressman G., 1959: An operational objective analysis system. Monthly
    Weather Review, 87(10), 367-374.

    Pauley, P. M. and X. Wu, 1990: The theoretical, discrete, and actual
    response of the Barnes objective analysis scheme for one- and
    two-dimensional fields. Monthly Weather Review, 118, 1145-1164

    """
    # make a tuple if passed a radar object as the first argument
    if isinstance(radars, Radar):
        radars = (radars, )

    # map the radar(s) to a cartesian grid
    if gridding_algo == 'map_to_grid':
        grids = map_to_grid(radars, grid_shape, grid_limits, **kwargs)
    elif gridding_algo == 'map_gates_to_grid':
        grids = map_gates_to_grid(radars, grid_shape, grid_limits, **kwargs)
    else:
        raise ValueError('invalid gridding_algo')

    # create and populate the field dictionary
    fields = {}
    first_radar = radars[0]

    for field in grids.keys():
        if field == 'ROI':
            fields['ROI'] = {
                'data': grids['ROI'],
                'standard_name': 'radius_of_influence',
                'long_name': 'Radius of influence for mapping',
                'units': 'm',
                'least_significant_digit': 1,
                '_FillValue': get_fillvalue()}
        else:
            fields[field] = {'data': grids[field]}
            # copy the metadata from the radar to the grid
            for key in first_radar.fields[field].keys():
                if key == 'data':
                    continue
                fields[field][key] = first_radar.fields[field][key]

    # time dictionaries
    time = get_metadata('grid_time')
    time['data'] = np.array([first_radar.time['data'][0]])
    time['units'] = first_radar.time['units']

    # grid coordinate dictionaries
    nz, ny, nx = grid_shape
    (z0, z1), (y0, y1), (x0, x1) = grid_limits

    x = get_metadata('x')
    x['data'] = np.linspace(x0, x1, nx)

    y = get_metadata('y')
    y['data'] = np.linspace(y0, y1, ny)

    z = get_metadata('z')
    z['data'] = np.linspace(z0, z1, nz)

    # grid origin location dictionaries
    origin_latitude = get_metadata('origin_latitude')
    origin_longitude = get_metadata('origin_longitude')
    if 'grid_origin' in kwargs:
        origin_latitude['data'] = np.array([kwargs['grid_origin'][0]])
        origin_longitude['data'] = np.array([kwargs['grid_origin'][1]])
    else:
        origin_latitude['data'] = first_radar.latitude['data']
        origin_longitude['data'] = first_radar.longitude['data']

    origin_altitude = get_metadata('origin_altitude')
    if 'grid_origin_alt' in kwargs:
        origin_altitude['data'] = np.array([kwargs['grid_origin_alt']])
    else:
        origin_altitude['data'] = first_radar.altitude['data']

    # metadata dictionary
    metadata = dict(first_radar.metadata)

    # create radar_ dictionaries
    radar_latitude = get_metadata('radar_latitude')
    radar_latitude['data'] = np.array(
        [r.latitude['data'][0] for r in radars])

    radar_longitude = get_metadata('radar_longitude')
    radar_longitude['data'] = np.array(
        [radar.longitude['data'][0] for radar in radars])

    radar_altitude = get_metadata('radar_altitude')
    radar_altitude['data'] = np.array(
        [radar.altitude['data'][0] for radar in radars])

    radar_time = get_metadata('radar_time')
    times, units = _unify_times_for_radars(radars)
    radar_time['units'] = units
    radar_time['data'] = times

    radar_name = get_metadata('radar_name')
    name_key = 'instrument_name'
    names = [radar.metadata[name_key] if name_key in radar.metadata else ''
             for radar in radars]
    radar_name['data'] = np.array(names)

    projection = kwargs.pop('grid_projection', None)

    return Grid(
        time, fields, metadata,
        origin_latitude, origin_longitude, origin_altitude, x, y, z,
        radar_latitude=radar_latitude, radar_longitude=radar_longitude,
        radar_altitude=radar_altitude, radar_name=radar_name,
        radar_time=radar_time, projection=projection)


def _unify_times_for_radars(radars):
    """ Return unified start times and units for a number of radars. """
    dates = [netCDF4.num2date(radar.time['data'][0], radar.time['units'])
             for radar in radars]
    units = make_time_unit_str(min(dates))
    times = netCDF4.date2num(dates, units)
    return times, units


class NNLocator:
    """
    Nearest neighbor locator.

    Class for finding the neighbors of a points within a given distance.

    Parameters
    ----------
    data : array_like, (n_sample, n_dimensions)
        Locations of points to be indexed. Note that if data is a
        C-contiguous array of dtype float64 the data will not be copied.
        Othersize and internal copy will be made.
    leafsize : int
        The number of points at which the algorithm switches over to
        brute-force. This can significantly impact the speed of the
        contruction and query of the tree.
    algorithm : 'kd_tree', optional.
        Algorithm used to compute the nearest neigbors. 'kd_tree' uses a
        k-d tree.

    """

    def __init__(self, data, leafsize=10, algorithm='kd_tree'):
        """ initalize. """
        self._algorithm = algorithm

        if algorithm == 'kd_tree':
            self.tree = cKDTree(data, leafsize=leafsize)
        else:
            raise ValueError('invalid algorithm')

    def find_neighbors_and_dists(self, q, r):
        """
        Find all neighbors and distances within a given distance.

        Parameters
        ----------
        q : n_dimensional tuple
            Point to query
        r : float
            Distance within which neighbors are returned.

        Returns
        -------
        ind : array of intergers
            Indices of the neighbors.
        dist : array of floats
            Distances to the neighbors.

        """
        if self._algorithm == 'kd_tree':
            ind = self.tree.query_ball_point(q, r)
            if np.size(ind) == 0:
                return ind, 0
            dist = scipy.spatial.minkowski_distance(q, self.tree.data[ind])

            return ind, dist


def map_to_grid(radars, grid_shape, grid_limits, grid_origin=None,
                grid_origin_alt=None, grid_projection=None,
                fields=None, gatefilters=False,
                map_roi=True, weighting_function='Barnes', toa=17000.0,
                copy_field_data=True, algorithm='kd_tree', leafsize=10.,
                roi_func='dist_beam', constant_roi=None,
                z_factor=0.05, xy_factor=0.02, min_radius=500.0,
                h_factor=1.0, nb=1.5, bsp=1.0, **kwargs):
    """
    Map one or more radars to a Cartesian grid.

    Generate a Cartesian grid of points for the requested fields from the
    collected points from one or more radars. The field value for a grid
    point is found by interpolating from the collected points within a given
    radius of influence and weighting these nearby points according to their
    distance from the grid points. Collected points are filtered
    according to a number of criteria so that undesired points are not
    included in the interpolation.

    Parameters
    ----------
    radars : Radar or tuple of Radar objects.
        Radar objects which will be mapped to the Cartesian grid.
    grid_shape : 3-tuple of floats
        Number of points in the grid (z, y, x).
    grid_limits : 3-tuple of 2-tuples
        Minimum and maximum grid location (inclusive) in meters for the
        z, y, x coordinates.
    grid_origin : (float, float) or None
        Latitude and longitude of grid origin. None sets the origin
        to the location of the first radar.
    grid_origin_alt: float or None
        Altitude of grid origin, in meters. None sets the origin
        to the location of the first radar.
    grid_projection : dic or str
        Projection parameters defining the map projection used to transform the
        locations of the radar gates in geographic coordinate to Cartesian
        coodinates. None will use the default dictionary which uses a native
        azimutal equidistance projection. See :py:func:`pyart.core.Grid` for
        additional details on this parameter. The geographic coordinates of
        the radar gates are calculated using the projection defined for each
        radar. No transformation is used if a grid_origin and grid_origin_alt
        are None and a single radar is specified.
    fields : list or None
        List of fields within the radar objects which will be mapped to
        the cartesian grid. None, the default, will map the fields which are
        present in all the radar objects.
    gatefilters : GateFilter, tuple of GateFilter objects, optional
        Specify what gates from each radar will be included in the
        interpolation onto the grid. Only gates specified in each gatefilters
        will be included in the mapping to the grid. A single GateFilter can
        be used if a single Radar is being mapped. A value of False for a
        specific element or the entire parameter will apply no filtering of
        gates for a specific radar or all radars (the default).
        Similarily a value of None will create a GateFilter from the
        radar moments using any additional arguments by passing them to
        :py:func:`moment_based_gate_filter`.
    roi_func : str or function
        Radius of influence function. A functions which takes an
        z, y, x grid location, in meters, and returns a radius (in meters)
        within which all collected points will be included in the weighting
        for that grid points. Examples can be found in the
        :py:func:`example_roi_func_constant`,
        :py:func:`example_roi_func_dist`, and
        :py:func:`example_roi_func_dist_beam`.
        Alternatively the following strings can use to specify a built in
        radius of influence function:

            * constant: constant radius of influence.
            * dist: radius grows with the distance from each radar.
            * dist_beam: radius grows with the distance from each radar
              and parameter are based of virtual beam sizes.

        The parameters which control these functions are listed in the
        `Other Parameters` section below.
    map_roi : bool
        True to include a radius of influence field in the returned
        dictionary under the 'ROI' key. This is the value of roi_func at all
        grid points.
    weighting_function : 'Barnes' or 'Barnes2' or 'Cressman' or 'Nearest'
        Functions used to weight nearby collected points when interpolating a
        grid point.
    toa : float
        Top of atmosphere in meters. Collected points above this height are
        not included in the interpolation.

    Other Parameters
    ----------------
    constant_roi : float
        Radius of influence parameter for the built in 'constant' function.
        This parameter is the constant radius in meter for all grid points.
        This parameter is used when `roi_func` is `constant` or constant_roi
        is not None. If constant_roi is not None, the constant roi_func is
        used automatically.
    z_factor, xy_factor, min_radius : float
        Radius of influence parameters for the built in 'dist' function.
        The parameter correspond to the radius size increase, in meters,
        per meter increase in the z-dimension from the nearest radar,
        the same foreach meteter in the xy-distance from the nearest radar,
        and the minimum radius of influence in meters. These parameters are
        only used when `roi_func` is 'dist'.
    h_factor, nb, bsp, min_radius : float
        Radius of influence parameters for the built in 'dist_beam' function.
        The parameter correspond to the height scaling, virtual beam width,
        virtual beam spacing, and minimum radius of influence. These
        parameters are only used when `roi_func` is 'dist_mean'.
    copy_field_data : bool
        True to copy the data within the radar fields for faster gridding,
        the dtype for all fields in the grid will be float64. False will not
        copy the data which preserves the dtype of the fields in the grid,
        may use less memory but results in significantly slower gridding
        times. When False gates which are masked in a particular field but
        are not masked in the `refl_field` field will still be included in
        the interpolation. This can be prevented by setting this parameter
        to True or by gridding each field individually setting the
        `refl_field` parameter and the `fields` parameter to the field in
        question. It is recommended to set this parameter to True.
    algorithm : 'kd_tree'.
        Algorithms to use for finding the nearest neighbors. 'kd_tree' is the
        only valid option.
    leafsize : int
        Leaf size passed to the neighbor lookup tree. This can affect the
        speed of the construction and query, as well as the memory required
        to store the tree. The optimal value depends on the nature of the
        problem. This value should only effect the speed of the gridding,
        not the results.

    Returns
    -------
    grids : dict
        Dictionary of mapped fields. The keys of the dictionary are given by
        parameter fields. Each elements is a `grid_size` float64 array
        containing the interpolated grid for that field.

    See Also
    --------
    grid_from_radars : Map to grid and return a Grid object.

    """
    # make a tuple if passed a radar object as the first argument
    if isinstance(radars, Radar):
        radars = (radars, )

    skip_transform = False
    if len(radars) == 1 and grid_origin_alt is None and grid_origin is None:
        skip_transform = True

    # parse the gatefilters argument
    if isinstance(gatefilters, GateFilter):
        gatefilters = (gatefilters, )  # make tuple if single filter passed
    if gatefilters is False:
        gatefilters = (False, ) * len(radars)
    if gatefilters is None:
        gatefilters = (None, ) * len(radars)
    if len(gatefilters) != len(radars):
        raise ValueError('Length of gatefilters must match length of radars')

    # check the parameters
    if weighting_function.upper() not in [
            'CRESSMAN', 'BARNES2', 'BARNES', 'NEAREST']:
        raise ValueError('unknown weighting_function ' +
                         weighting_function.upper())

    if algorithm not in ['kd_tree']:
        raise ValueError('unknown algorithm: %s' % algorithm)
    badval = get_fillvalue()

    # parse the grid_projection
    if grid_projection is None:
        grid_projection = {
            'proj': 'pyart_aeqd', '_include_lon_0_lat_0': True}

    # find the grid origin if not given
    if grid_origin is None:
        try:
            lat = float(radars[0].latitude['data'])
            lon = float(radars[0].longitude['data'])
        except TypeError:
            lat = np.mean(radars[0].latitude['data'])
            lon = np.mean(radars[0].longitude['data'])
        grid_origin = (lat, lon)
    grid_origin_lat, grid_origin_lon = grid_origin

    if grid_origin_alt is None:
        try:
            grid_origin_alt = float(radars[0].altitude['data'])
        except TypeError:
            grid_origin_alt = np.mean(radars[0].altitude['data'])

    # fields which should be mapped, None for fields which are in all radars
    if fields is None:
        fields = set(radars[0].fields.keys())
        for radar in radars[1:]:
            fields = fields.intersection(radar.fields.keys())
        fields = list(fields)
    nfields = len(fields)

    # determine the number of gates (collected points) in each radar
    nradars = len(radars)
    ngates_per_radar = [r.fields[fields[0]]['data'].size for r in radars]
    total_gates = np.sum(ngates_per_radar)
    gate_offset = np.cumsum([0] + ngates_per_radar)

    # create arrays to hold the gate locations and indicators if the gate
    # should be included in the interpolation.
    gate_locations = np.empty((total_gates, 3), dtype=np.float64)
    include_gate = np.ones((total_gates), dtype=np.bool)

    offsets = []    # offsets from the grid origin, in meters, for each radar

    # create a field lookup tables
    if copy_field_data:
        # copy_field_data == True, lookups are performed on a 2D copy of
        # all of the field data in all radar objects, this can be a
        # large array.  These lookup are fast as the dtype is know.
        field_data = np.ma.empty((total_gates, nfields), dtype=np.float64)
    else:
        # copy_field_data == False, lookups are performed on a 2D object
        # array pointing to the radar fields themselved, no copies are made.
        # A table mapping filtered gates to raw gates is created later.
        # Since the dtype is not not know this method is slow.
        field_data_objs = np.empty((nfields, nradars), dtype='object')
        # We also need to know how many gates from each radar are included
        # in the NNLocator, the filtered_gates_per_radar list records this
        filtered_gates_per_radar = []

    projparams = grid_projection.copy()
    if projparams.pop('_include_lon_0_lat_0', False):
        projparams['lon_0'] = grid_origin_lon
        projparams['lat_0'] = grid_origin_lat

    # loop over the radars finding gate locations, field data, and offset
    for iradar, (radar, gatefilter) in enumerate(zip(radars, gatefilters)):

        # calculate radar offset from the origin
        x_disp, y_disp = geographic_to_cartesian(
            radar.longitude['data'], radar.latitude['data'], projparams)
        try:
            z_disp = float(radar.altitude['data']) - grid_origin_alt
            offsets.append((z_disp, float(y_disp), float(x_disp)))
        except TypeError:
            z_disp = np.mean(radar.altitude['data']) - grid_origin_alt
            offsets.append((z_disp, np.mean(y_disp), np.mean(x_disp)))

        # calculate cartesian locations of gates
        if skip_transform:
            xg_loc = radar.gate_x['data']
            yg_loc = radar.gate_y['data']
        else:
            xg_loc, yg_loc = geographic_to_cartesian(
                radar.gate_longitude['data'], radar.gate_latitude['data'],
                projparams)
        zg_loc = radar.gate_altitude['data'] - grid_origin_alt

        # add gate locations to gate_locations array
        start, end = gate_offset[iradar], gate_offset[iradar + 1]
        gate_locations[start:end, 0] = zg_loc.flat[:]
        gate_locations[start:end, 1] = yg_loc.flat[:]
        gate_locations[start:end, 2] = xg_loc.flat[:]

        del xg_loc, yg_loc

        # determine which gates should be included in the interpolation
        gflags = zg_loc < toa      # include only those below toa

        if gatefilter is not False:
            # excluded gates marked by the gatefilter
            if gatefilter is None:
                gatefilter = moment_based_gate_filter(radar, **kwargs)
            gflags = np.logical_and(gflags, gatefilter.gate_included)

        include_gate[start:end] = gflags.ravel()

        if not copy_field_data:
            # record the number of gates from the current radar which
            # are included in the interpolation.
            filtered_gates_per_radar.append(gflags.sum())

        del gflags, zg_loc

        # copy/store references to field data for lookup
        for ifield, field in enumerate(fields):
            flat_field_data = radar.fields[field]['data'].ravel()
            if copy_field_data:
                field_data[start:end, ifield] = flat_field_data
            else:
                field_data_objs[ifield, iradar] = flat_field_data
        del flat_field_data

    # build field data lookup tables
    if copy_field_data:
        # copy_field_data == True we filtered the field data in the
        # same manner as we will filter the gate locations.
        filtered_field_data = field_data[include_gate]
    else:
        # copy_field_data == True, build a lookup table which maps from
        # filtered gate number to (radar number, radar gate number)
        # the radar number is given as the quotent of the lookup table
        # value divided by total_gates, the remainer gives the index of
        # the gate in the flat field data array.

        # initalize the lookup table with values from 0 ... total gates
        lookup = np.where(include_gate)[0]

        # number of filtered gates before a given radar
        filtered_gate_offset = np.cumsum([0] + filtered_gates_per_radar)

        # for radars 1 to N-1 add total_gates to the lookup table and
        # subtract the number of gates in all ealier radars.
        for i in range(1, nradars):
            l_start = filtered_gate_offset[i]
            l_end = filtered_gate_offset[i + 1]
            gates_before = gate_offset[i]
            lookup[l_start:l_end] += (total_gates * i - gates_before)

    # populate the nearest neighbor locator with the filtered gate locations
    nnlocator = NNLocator(gate_locations[include_gate], algorithm=algorithm,
                          leafsize=leafsize)

    # unpack the grid parameters
    nz, ny, nx = grid_shape
    zr, yr, xr = grid_limits
    z_start, z_stop = zr
    y_start, y_stop = yr
    x_start, x_stop = xr

    if nz == 1:
        z_step = 0.
    else:
        z_step = (z_stop - z_start) / (nz - 1.)
    if ny == 1:
        y_step = 0.
    else:
        y_step = (y_stop - y_start) / (ny - 1.)
    if nx == 1:
        x_step = 0.
    else:
        x_step = (x_stop - x_start) / (nx - 1.)

    if not hasattr(roi_func, '__call__'):
        if constant_roi is not None:
            roi_func = 'constant'
        else:
            constant_roi = 500.0
        if roi_func == 'constant':
            roi_func = _gen_roi_func_constant(constant_roi)
        elif roi_func == 'dist':
            roi_func = _gen_roi_func_dist(
                z_factor, xy_factor, min_radius, offsets)
        elif roi_func == 'dist_beam':
            roi_func = _gen_roi_func_dist_beam(
                h_factor, nb, bsp, min_radius, offsets)
        else:
            raise ValueError('unknown roi_func: %s' % roi_func)

    # create array to hold interpolated grid data and roi if requested
    grid_data = np.ma.empty((nz, ny, nx, nfields), dtype=np.float64)
    grid_data.set_fill_value(badval)

    if map_roi:
        roi = np.empty((nz, ny, nx), dtype=np.float64)

    # interpolate field values for each point in the grid
    for iz, iy, ix in np.ndindex(nz, ny, nx):

        # calculate the grid point
        x = x_start + x_step * ix
        y = y_start + y_step * iy
        z = z_start + z_step * iz
        r = roi_func(z, y, x)
        if map_roi:
            roi[iz, iy, ix] = r

        # find neighbors and distances
        ind, dist = nnlocator.find_neighbors_and_dists((z, y, x), r)

        if np.size(ind) == 0:
            # when there are no neighbors, mark the grid point as bad
            grid_data[iz, iy, ix] = np.ma.masked
            grid_data.data[iz, iy, ix] = badval
            continue

        # find the field values for all neighbors
        if copy_field_data:
            # copy_field_data == True, a slice will get the field data.
            nn_field_data = filtered_field_data[ind]
        else:
            # copy_field_data == False, use the lookup table to find the
            # radar numbers and gate numbers for the neighbors. Then
            # use the _load_nn_field_data function to load this data from
            # the field data object array. This is done in Cython for speed.
            r_nums, e_nums = divmod(lookup[ind], total_gates)
            npoints = r_nums.size
            r_nums = r_nums.astype(np.intc)
            e_nums = e_nums.astype(np.intc)
            nn_field_data = np.empty((npoints, nfields), np.float64)
            _load_nn_field_data(field_data_objs, nfields, npoints, r_nums,
                                e_nums, nn_field_data)

        # preforms weighting of neighbors.
        dist2 = dist * dist
        r2 = r * r

        if weighting_function.upper() == 'NEAREST':
            value = nn_field_data[np.argmin(dist2)]
        else:
            if weighting_function.upper() == 'CRESSMAN':
                weights = (r2 - dist2) / (r2 + dist2)
            elif weighting_function.upper() == 'BARNES':
                warnings.warn("Barnes weighting function is deprecated."
                              " Please use Barnes 2 to be consistent with"
                              " Pauley and Wu 1990.", DeprecationWarning)
                weights = np.exp(-dist2 / (2.0 * r2)) + 1e-5
            elif weighting_function.upper() == 'BARNES2':
                weights = np.exp(-dist2 / (r2/4)) + 1e-5
            value = np.ma.average(nn_field_data, weights=weights, axis=0)

        grid_data[iz, iy, ix] = value

    # create and return the grid dictionary
    grids = dict([(f, grid_data[..., i]) for i, f in enumerate(fields)])
    if map_roi:
        grids['ROI'] = roi
    return grids


# Radius of Influence (RoI) functions


def example_roi_func_constant(zg, yg, xg):
    """
    Example RoI function which returns a constant radius.

    Parameters
    ----------
    zg, yg, xg : float
        Distance from the grid center in meters for the x, y and z axes.

    Returns
    -------
    roi : float
        Radius of influence in meters
    """
    # RoI function parameters
    constant = 500.     # constant 500 meter RoI
    return constant


def _gen_roi_func_constant(constant_roi):
    """
    Return a RoI function which returns a constant radius.

    See :py:func:`map_to_grid` for a description of the parameters.
    """

    def roi(zg, yg, xg):
        """ constant radius of influence function. """
        return constant_roi

    return roi


def example_roi_func_dist(zg, yg, xg):
    """
    Example RoI function which returns a radius which grows with distance.

    Parameters
    ----------
    zg, yg, xg : float
        Distance from the grid center in meters for the x, y and z axes.

    Returns
    -------
    roi : float

    """
    # RoI function parameters
    z_factor = 0.05         # increase in radius per meter increase in z dim
    xy_factor = 0.02        # increase in radius per meter increase in xy dim
    min_radius = 500.       # minimum radius
    offsets = ((0, 0, 0), )  # z, y, x offset of grid in meters from radar(s)

    offsets = np.array(offsets)
    zg_off = offsets[:, 0]
    yg_off = offsets[:, 1]
    xg_off = offsets[:, 2]
    r = np.maximum(z_factor * (zg - zg_off) +
                   xy_factor * np.sqrt((xg - xg_off)**2 + (yg - yg_off)**2),
                   min_radius)
    return min(r)


def _gen_roi_func_dist(z_factor, xy_factor, min_radius, offsets):
    """
    Return a RoI function whose radius grows with distance.

    See :py:func:`map_to_grid` for a description of the parameters.
    """
    offsets = np.array(offsets)
    zg_off = offsets[:, 0]
    yg_off = offsets[:, 1]
    xg_off = offsets[:, 2]

    def roi(zg, yg, xg):
        """ dist radius of influence function. """
        r = np.maximum(
            z_factor * (zg - zg_off) +
            xy_factor * np.sqrt((xg - xg_off)**2 + (yg - yg_off)**2),
            min_radius)
        return min(r)

    return roi


def example_roi_func_dist_beam(zg, yg, xg):
    """
    Example RoI function which returns a radius which grows with distance
    and whose parameters are based on virtual beam size.

    Parameters
    ----------
    zg, yg, xg : float
        Distance from the grid center in meters for the x, y and z axes.

    Returns
    -------
    roi : float

    """
    # RoI function parameters
    h_factor = 1.0      # height scaling
    nb = 1.5            # virtual beam width
    bsp = 1.0           # virtual beam spacing
    min_radius = 500.   # minimum radius in meters
    offsets = ((0, 0, 0), )  # z, y, x offset of grid in meters from radar(s)

    offsets = np.array(offsets)
    zg_off = offsets[:, 0]
    yg_off = offsets[:, 1]
    xg_off = offsets[:, 2]
    r = np.maximum(
        h_factor * ((zg - zg_off) / 20.0) +
        np.sqrt((yg - yg_off)**2 + (xg - xg_off)**2) *
        np.tan(nb * bsp * np.pi / 180.0), min_radius)
    return min(r)


def _gen_roi_func_dist_beam(h_factor, nb, bsp, min_radius, offsets):
    """
    Return a RoI function whose radius which grows with distance
    and whose parameters are based on virtual beam size.

    See :py:func:`map_to_grid` for a description of the parameters.
    """
    offsets = np.array(offsets)
    zg_off = offsets[:, 0]
    yg_off = offsets[:, 1]
    xg_off = offsets[:, 2]

    def roi(zg, yg, xg):
        """ dist_beam radius of influence function. """
        r = np.maximum(
            h_factor * ((zg - zg_off) / 20.0) +
            np.sqrt((yg - yg_off)**2 + (xg - xg_off)**2) *
            np.tan(nb * bsp * np.pi / 180.0), min_radius)
        return min(r)

    return roi



"""
pyart.correct._common_dealias
=============================

Routines used by multiple dealiasing functions.

.. autosummary::
    :toctree: generated/

    _parse_fields
    _parse_nyquist_vel
    _parse_gatefilter
    _parse_rays_wrap_around
    _set_limits

"""

import numpy as np

from ..config import get_field_name
from ..filters.gatefilter import moment_based_gate_filter, GateFilter


def _parse_fields(vel_field, corr_vel_field):
    """ Parse and return the radar fields for dealiasing. """
    if vel_field is None:
        vel_field = get_field_name('velocity')
    if corr_vel_field is None:
        corr_vel_field = get_field_name('corrected_velocity')
    return vel_field, corr_vel_field


def _parse_nyquist_vel(nyquist_vel, radar, check_uniform):
    """ Parse the nyquist_vel parameter, extract from the radar if needed. """
    if nyquist_vel is None:
        nyquist_vel = [radar.get_nyquist_vel(i, check_uniform) for i in
                       range(radar.nsweeps)]
    else:   # Nyquist velocity explicitly provided
        try:
            len(nyquist_vel)
        except:   # expand single value.
            nyquist_vel = [nyquist_vel for i in range(radar.nsweeps)]
    return nyquist_vel


def _parse_gatefilter(gatefilter, radar, **kwargs):
    """ Parse the gatefilter, return a valid GateFilter object. """
    # parse the gatefilter parameter
    if gatefilter is None:  # create a moment based filter
        gatefilter = moment_based_gate_filter(radar, **kwargs)
    elif gatefilter is False:
        gatefilter = GateFilter(radar)
    else:
        gatefilter = gatefilter.copy()
    return gatefilter


def _parse_rays_wrap_around(rays_wrap_around, radar):
    """ Parse the rays_wrap_around parameter. """
    if rays_wrap_around is None:
        if radar.scan_type == 'ppi':
            rays_wrap_around = True
        else:
            rays_wrap_around = False
    return rays_wrap_around


def _set_limits(data, nyquist_vel, dic):
    """ Set the valid_min and valid_max keys in dic from dealiased data. """
    max_abs_vel = np.ma.max(np.ma.abs(data))
    if max_abs_vel is np.ma.masked:
        # all velocities are masked, do not set valid_min and valid_max
        return
    max_nyq_vel = np.ma.max(nyquist_vel)
    max_nyq_int = 2. * max_nyq_vel
    added_intervals = np.ceil((max_abs_vel - max_nyq_vel) / (max_nyq_int))
    max_valid_velocity = (max_nyq_vel + added_intervals * max_nyq_int)
    dic['valid_min'] = float(-max_valid_velocity)
    dic['valid_max'] = float(max_valid_velocity)
    return



"""
pyart.io.chl
============

Utilities for reading CSU-CHILL CHL files.

.. autosummary::
    :toctree: generated/
    :template: dev_template.rst

    ChlFile

.. autosummary::
    :toctree: generated/

    read_chl
    _unpack_structure

"""

from datetime import datetime
import struct

import numpy as np

from ..config import FileMetadata, get_fillvalue
from ..core.radar import Radar
from .common import make_time_unit_str, _test_arguments, prepare_for_read


def read_chl(filename, field_names=None, additional_metadata=None,
             file_field_names=None, exclude_fields=None,
             include_fields=None, use_file_field_attributes=True, **kwargs):
    """
    Read a CSU-CHILL CHL file.

    Parameters
    ----------
    filename : str
        Name of CHL file.
    field_names : dict, optional
        Dictionary mapping CHL field names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        Py-ART configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata from during this read.
        This metadata is not used during any successive file reads unless
        explicitly included. A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    file_field_names : bool, optional
        True to use the CHL field names for the field names in the radar
        object. If this case the field_names parameter is ignored.
        The field dictionary will likely only have a 'data' key, unless
        the fields are defined in `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `field_file_names` and `field_names` parameters. Set to
        None to include all fields not in exclude_fields.
    use_file_field_attributes : bool, optional
        True to use information provided by in the file to set the field
        attribute `long_name`, `units`, `valid_max`, and `valid_min`. False
        will not set these unless they are defined in the configuration file
        or in `additional_metadata`.

    Returns
    -------
    radar : Radar
        Radar object containing data from CHL file.

    """
    # test for non empty kwargs
    _test_arguments(kwargs)

    # create metadata retrival object
    filemetadata = FileMetadata('chl', field_names, additional_metadata,
                                file_field_names, exclude_fields,
                                include_fields)

    # read data
    chl_file = ChlFile(prepare_for_read(filename))

    # time
    time = filemetadata('time')
    tdata = np.array(chl_file.time)
    min_time = np.floor(tdata.min())
    time['data'] = (tdata - min_time).astype('float64')
    time['units'] = make_time_unit_str(datetime.utcfromtimestamp(min_time))

    # range
    _range = filemetadata('range')
    _range['data'] = (np.array(range(chl_file.ngates)) *
                      chl_file.gate_spacing + chl_file.first_gate_offset)
    _range['meters_between_gates'] = np.array(chl_file.gate_spacing)
    _range['meters_to_center_of_first_gate'] = 0.0

    # scan_type
    scan_type = SCAN_MODE_NAMES[chl_file.scan_types[-1]]

    # fields
    fields = {}
    for i, fdata in chl_file.fields.items():

        field_info = chl_file.field_info[i]
        field_name = filemetadata.get_field_name(field_info['name'])
        if field_name is None:
            continue

        field_dic = filemetadata(field_name)
        np.ma.set_fill_value(fdata, get_fillvalue())
        field_dic['data'] = fdata
        field_dic['_FillValue'] = get_fillvalue()

        if use_file_field_attributes:
            field_dic['long_name'] = field_info['descr']
            field_dic['units'] = field_info['units']
            field_dic['valid_max'] = field_info['max_val']
            field_dic['valid_min'] = field_info['min_val']

        fields[field_name] = field_dic

    # metadata
    metadata = filemetadata('metadata')
    metadata['instrument_name'] = chl_file.radar_info['radar_name']
    metadata['original_container'] = 'CHL'

    # longitude, latitude, altitude
    latitude = filemetadata('latitude')
    longitude = filemetadata('longitude')
    altitude = filemetadata('altitude')
    latitude['data'] = np.array([chl_file.radar_info['latitude']], 'f8')
    longitude['data'] = np.array([chl_file.radar_info['longitude']], 'f8')
    altitude['data'] = np.array([chl_file.radar_info['altitude']], 'f8')

    # sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
    # sweep_end_ray_index
    sweep_number = filemetadata('sweep_number')
    sweep_mode = filemetadata('sweep_mode')
    fixed_angle = filemetadata('fixed_angle')
    sweep_start_ray_index = filemetadata('sweep_start_ray_index')
    sweep_end_ray_index = filemetadata('sweep_end_ray_index')
    sweep_number['data'] = np.arange(chl_file.num_sweeps, dtype='int32')
    sweep_mode['data'] = np.array(
        [SCAN_MODE_NAMES[i] for i in chl_file.scan_types], dtype='S')
    fixed_angle['data'] = np.array(chl_file.fixed_angle, dtype='float32')

    ray_count = chl_file.rays_per_sweep
    ssri = np.cumsum(np.append([0], ray_count[:-1])).astype('int32')
    sweep_start_ray_index['data'] = ssri
    sweep_end_ray_index['data'] = np.cumsum(ray_count).astype('int32') - 1

    # azimuth, elevation
    azimuth = filemetadata('azimuth')
    elevation = filemetadata('elevation')
    azimuth['data'] = np.array(chl_file.azimuth, dtype='float32')
    elevation['data'] = np.array(chl_file.elevation, dtype='float32')

    # instrument parameters
    instrument_parameters = None

    chl_file.close()
    return Radar(
        time, _range, fields, metadata, scan_type,
        latitude, longitude, altitude,
        sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth, elevation,
        instrument_parameters=instrument_parameters)


class ChlFile(object):
    """
    A file object for CHL data.

    Parameters
    ----------
    filename : str or file-like.
        Name of CHL file to read or a file-like object pointing to the
        beginning of such a file.
    ns_time : bool
        True to determine ray collection times to the nano-second, False
        will only determine times to the second.
    debug : bool
        True to keep packet data in the _packets attribute to aid in
        debugging.

    Attributes
    ----------
    ngates : int
        Number of gates per ray.
    num_sweeps : int
        Number of sweeps in the volume.
    gate_spacing : float
        Spacing in meters between gates.
    first_gate_offset : float
        Distance in meters to the first range gate.
    time : list of ints
        Time in seconds in epoch for each ray in the volume.
    azimuth : list of floats
        Azimuth angle for each ray in the volume in degrees.
    elevation : list of floats
        Elevation angle for each ray in the volume in degrees.
    fixed_angle : list of floats
        Fixed angles for each sweep.
    sweep_number : list of ints
        Sweep numbers reported in file.
    scan_types : list of ints
        Chill defined scan type for each sweep.
    rays_per_sweep : list of ints
        Number of rays in each sweep.
    fields : dict
        Dictionary of field data index by field number.
    radar_info : dict
        Radar information recorded in the file.
    field_info : dict
        Field information (limits, name, etc.) recorded in the file.
    processor_info : dict
        Processor information recorded in the file.

    """

    def __init__(self, filename, ns_time=True, debug=False):

        # initalize attributes
        self.ngates = None
        self.num_sweeps = None
        self.gate_spacing = None
        self.time = []
        self.azimuth = []
        self.elevation = []
        self.fixed_angle = []
        self.sweep_number = []
        self.scan_types = []
        self.rays_per_sweep = None
        self.fields = {}
        self.radar_info = None
        self.field_info = {}
        self.processor_info = None
        self.first_gate_offset = None

        # private attributes
        self._dstring = b''     # string containing field data.
        self._bit_mask = None   # bit mask specifying fields present in file.
        self._dtype = None      # NumPy dtype for a single gate (all fields).
        self._ray_bsize = None  # size in bytes of a single ray (all fields).
        self._packets = []      # List of packets, not set if debug is False.
        self._field_nums = None     # Field number available in file.
        self._rays_in_current_sweep = None  # accumulator for counting rays.
        self._fh = None         # file handler
        self._include_ns_time = ns_time

        # read all blocks from the file
        if hasattr(filename, 'read'):
            self._fh = filename
        else:
            self._fh = open(filename, "rb")
        packet = 1
        while packet is not None:
            packet = self._read_block()
            if debug:
                self._packets.append(packet)

        self._extract_fields()
        self.rays_per_sweep.append(self._rays_in_current_sweep)

    def close(self):
        """ Close the file. """
        self._fh.close()

    def _read_block(self):
        """ Read a block from an open CHL file. """
        pld = self._fh.read(8)
        if pld == b'':
            return None
        block_id, length = struct.unpack("<2i", pld)
        payload = self._fh.read(length - 8)

        # parse the block
        if block_id == ARCH_ID_FILE_HDR:
            packet = self._parse_file_hdr_block(payload)
        elif block_id == ARCH_ID_FIELD_SCALE:
            packet = self._parse_field_scale_block(payload)
        elif block_id == ARCH_ID_RAY_HDR:
            packet = self._parse_ray_hdr_block(payload)
        elif block_id == HSK_ID_RADAR_INFO:
            packet = self._parse_radar_info_block(payload)
        elif block_id == HSK_ID_PROCESSOR_INFO:
            packet = self._parse_processor_info_block(payload)
        elif block_id == HSK_ID_SCAN_SEG:
            packet = self._parse_scan_seg_block(payload)
        elif block_id == ARCH_ID_SWEEP_BLOCK:
            packet = self._parse_sweep_block(payload)
        else:
            packet = {}
        packet['block_id'] = block_id
        packet['length'] = length
        return packet

    # Block parsers
    def _parse_file_hdr_block(self, payload):
        """ Parse a field_hdr block. """
        return _unpack_structure(payload, ARCH_FILE_HDR_T)

    def _parse_field_scale_block(self, payload):
        """ Parse a field_scale block. Add scale to field_info attr. """
        packet = _unpack_structure(payload, FIELD_SCALE_T)
        packet['name'] = packet['name'].decode('utf-8').rstrip('\x00')
        packet['units'] = packet['units'].decode('utf-8').rstrip('\x00')
        packet['descr'] = packet['descr'].decode('utf-8').rstrip('\x00')
        self.field_info[packet['bit_mask_pos']] = packet
        return packet

    def _parse_radar_info_block(self, payload):
        """ Parse a radar_info block. Update metadata attribute. """
        packet = _unpack_structure(payload, RADAR_INFO_T)
        packet['radar_name'] = (
            packet['radar_name'].decode('utf-8').rstrip('\x00'))
        self.radar_info = packet.copy()
        return packet

    def _parse_processor_info_block(self, payload):
        """ Parse a processor_info block. Set dr attribute. """
        packet = _unpack_structure(payload, PROCESSOR_INFO)
        self.gate_spacing = packet['gate_spacing']
        self.first_gate_offset = packet['range_offset']
        self.processor_info = packet.copy()
        return packet

    def _parse_scan_seg_block(self, payload):
        """ Parse a scan_seg_block. Update sweep attributes. """
        packet = _unpack_structure(payload, SCAN_SEG)
        self.sweep_number.append(packet['sweep_num'])
        self.fixed_angle.append(packet['current_fixed_angle'])
        self.scan_types.append(packet['scan_type'])
        if self.rays_per_sweep is None:
            self.rays_per_sweep = []    # first sweep block
            self._rays_in_current_sweep = 0
        else:
            self.rays_per_sweep.append(self._rays_in_current_sweep)
            self._rays_in_current_sweep = 0
        return packet

    def _parse_sweep_block(self, payload):
        """ Parse a sweep block. Set num_sweeps attribute. """
        packet = {}
        packet['num_sweeps'] = struct.unpack('I', payload[0:4])[0]
        packet['swp_offsets'] = struct.unpack(
            str(packet['num_sweeps']) + 'Q', payload[4:])
        self.num_sweeps = packet['num_sweeps']
        return packet

    def _parse_ray_hdr_block(self, payload):
        """ Parse a ray_hdr block. Update associated attributes. """
        packet = _unpack_structure(payload, ARCH_RAY_HEADER)

        if self._bit_mask is None:
            # this is the first ray_hdr block read
            self.ngates = packet['gates']
            self._bit_mask = packet['bit_mask']
            self._field_nums = [b for b in range(38) if self._bit_mask & 2**b]
            self._dtype = ','.join([DATA_FORMAT[self.field_info[i]['format']]
                                    for i in self._field_nums])
            self._ray_bsize = np.dtype(self._dtype).itemsize * packet['gates']
        else:
            # check that the bit_mask and number of gates are constant
            if packet['bit_mask'] != self._bit_mask:
                raise NotImplementedError('bit_mask is not consistent.')
            if packet['gates'] != self.ngates:
                raise NotImplementedError('number of gates vary.')

        # store ray data and pointing data
        self._dstring += self._fh.read(self._ray_bsize)
        if self._include_ns_time:
            self.time.append(packet['time'] + packet['ns_time']/1e9)
        else:
            self.time.append(packet['time'])
        self.azimuth.append(packet['azimuth'])
        self.elevation.append(packet['elevation'])
        self._rays_in_current_sweep += 1

        return packet

    def _extract_fields(self):
        """ Extract field data from _dstring attribute post read. """
        all_data = np.frombuffer(self._dstring, dtype=self._dtype)
        all_data = all_data.reshape(-1, self.ngates)
        for i, field_num in enumerate(self._field_nums):

            fdata = np.ma.masked_values(all_data[all_data.dtype.names[i]], 0)
            # apply scale and offset factors to interger data types
            if issubclass(fdata.dtype.type, np.integer):
                dat_factor = float(self.field_info[field_num]['dat_factor'])
                dat_bias = float(self.field_info[field_num]['dat_bias'])
                fld_factor = float(self.field_info[field_num]['fld_factor'])
                fdata = (fdata * dat_factor + dat_bias) / fld_factor

            self.fields[field_num] = fdata
        return

# CHL packet types
ARCH_FORMAT_VERSION = 0x00010000
ARCH_ID_CONTROL = 0x5aa80001
ARCH_ID_FIELD_SCALE = 0x5aa80002
ARCH_ID_RAY_HDR = 0x5aa80003
ARCH_ID_FILE_HDR = 0x5aa80004
ARCH_ID_SWEEP_BLOCK = 0x5aa80005
HSK_ID_PROCESSOR_INFO = 0x5aa50003
HSK_ID_RADAR_INFO = 0x5aa50001
HSK_ID_SCAN_SEG = 0x5aa50002

# Additional constants
SCAN_MODE_NAMES = ['ppi', 'rhi', 'fixed', 'manual ppi', 'manual rhi', 'idle']
DATA_FORMAT = ['uint8', 'uint64', 'float32', 'uint16']

##############
# Structures #
##############


def _unpack_structure(string, structure):
    """ Unpack a structure. """
    fmt = ''.join([i[1] for i in structure])
    tpl = struct.unpack(fmt, string)
    return dict(zip([i[0] for i in structure], tpl))

ARCH_FILE_HDR_T = (
    ('version', 'I'),
    ('creation_version', 'I'),
    ('creator_id', '32s'),
    ('sweep_table_offset', 'Q'),
)

ARCH_RAY_HEADER = (
    ('azimuth', 'f'),
    ('elevation', 'f'),
    ('azimuth_width', 'f'),
    ('elevation_width', 'f'),
    ('gates', 'H'),
    ('beam_index', 'H'),
    ('ns_time', 'I'),
    ('time', 'Q'),
    ('bit_mask', 'Q'),
    ('ray_number', 'I'),
    ('num_pulses', 'I'),
)

FIELD_SCALE_T = (
    ('format', 'i'),
    ('min_val', 'f'),
    ('max_val', 'f'),
    ('bit_mask_pos', 'i'),
    ('type_hint', 'i'),
    ('fld_factor', 'i'),
    ('dat_factor', 'i'),
    ('dat_bias', 'i'),
    ('name', '32s'),
    ('units', '32s'),
    ('descr', '128s')
)

RADAR_INFO_T = (
    ('radar_name', '32s'),
    ('latitude', 'f'),
    ('longitude', 'f'),
    ('altitude', 'f'),
    ('beamwidth', 'f'),
    ('wavelength_cm', 'f'),
    ('unused1', 'f'),
    ('unused2', 'f'),
    ('unused3', 'f'),
    ('unused4', 'f'),
    ('gain_ant_h', 'f'),
    ('gain_ant_v', 'f'),
    ('zdr_cal_base', 'f'),
    ('phidp_rot', 'f'),
    ('base_radar_constant', 'f'),
    ('unused5', 'f'),
    ('power_measurement_loss_h', 'f'),
    ('power_measurement_loss_v', 'f'),
    ('zdr_cal_base_vhs', 'f'),
    ('test_power_h', 'f'),
    ('test_power_v', 'f'),
    ('dc_loss_h', 'f'),
    ('dc_loss_v', 'f'),
)

PROCESSOR_INFO = (
    ('polarization_mode', 'i'),
    ('processing_mode', 'i'),
    ('pulse_type', 'i'),
    ('test_type', 'i'),
    ('integration_cycle_pulses', 'I'),
    ('clutter_filter_number', 'I'),
    ('range_gate_averaging', 'I'),
    ('indexed_beam_width', 'f'),
    ('gate_spacing', 'f'),
    ('prt_usec', 'f'),
    ('range_start', 'f'),
    ('range_stop', 'f'),
    ('max_gate', 'I'),
    ('test_power', 'f'),
    ('unused1', 'f'),
    ('unused2', 'f'),
    ('test_pulse_range', 'f'),
    ('test_pulse_length', 'f'),
    ('prt2', 'f'),
    ('range_offset', 'f'),
)

SCAN_SEG = (
    ('az_manual', 'f'),
    ('el_manual', 'f'),
    ('az_start', 'f'),
    ('el_start', 'f'),
    ('scan_rate', 'f'),
    ('segname', '24s'),
    ('opt', 'i'),
    ('follow_mode', 'i'),
    ('scan_type', 'i'),
    ('scan_flags', 'I'),
    ('volume_num', 'I'),
    ('sweep_num', 'I'),
    ('time_limit', 'I'),
    ('webtilt', 'I'),
    ('left_limit', 'f'),
    ('right_limit', 'f'),
    ('up_limit', 'f'),
    ('down_limit', 'f'),
    ('step', 'f'),
    ('max_sweeps', 'I'),
    ('filter_break_sweep', 'I'),
    ('clutter_filter1', 'I'),
    ('clutter_filter2', 'I'),
    ('project', '16s'),
    ('current_fixed_angle', 'f'),
)



"""
=======================================
Plot a cross section from  a PPI volume
=======================================

An example which extracts a cross section at two azimuth angles from a volume
of PPI scans and plots both cross sections.

"""
print(__doc__)

# Author: Jonathan J. Helmus (jhelmus@anl.gov)
# License: BSD 3 clause

import matplotlib.pyplot as plt
import pyart

filename = 'XSW110520105408.RAW7HHF'

radar = pyart.io.read(filename)
xsect = pyart.util.cross_section_ppi(radar, [45, 90])

display = pyart.graph.RadarDisplay(xsect)
fig = plt.figure()
ax = fig.add_subplot(211)
display.plot('reflectivity', 0, vmin=-32, vmax=64.)
ax = fig.add_subplot(212)
display.plot('reflectivity', 1, vmin=-32, vmax=64.)
plt.tight_layout()
plt.show()



"""
pyart.io.mdv_radar
==================

Utilities for reading of MDV radar files.

.. autosummary::
    :toctree: generated/

    read_mdv

"""

import numpy as np
from netCDF4 import date2num

from ..config import FileMetadata, get_fillvalue
from ..core.radar import Radar
from .common import make_time_unit_str, _test_arguments, prepare_for_read
from ..lazydict import LazyLoadDict
from . import mdv_common


def read_mdv(filename, field_names=None, additional_metadata=None,
             file_field_names=False, exclude_fields=None,
             include_fields=None, delay_field_loading=False, **kwargs):
    """
    Read a MDV file.

    Parameters
    ----------
    filename : str
        Name of MDV file to read or file-like object pointing to the
        beginning of such a file.
    field_names : dict, optional
        Dictionary mapping MDV data type names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        Py-ART configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata from during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    file_field_names : bool, optional
        True to use the MDV data type names for the field names. If this
        case the field_names parameter is ignored. The field dictionary will
        likely only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.
    delay_field_loading : bool
        True to delay loading of field data from the file until the 'data'
        key in a particular field dictionary is accessed. In this case
        the field attribute of the returned Radar object will contain
        LazyLoadDict objects not dict objects. Not all file types support this
        parameter.

    Returns
    -------
    radar : Radar
        Radar object containing data from MDV file.

    Notes
    -----
    Currently this function can only read polar MDV files with fields
    compressed with gzip or zlib.

    """
    # test for non empty kwargs
    _test_arguments(kwargs)

    # create metadata retrieval object
    filemetadata = FileMetadata('mdv', field_names, additional_metadata,
                                file_field_names, exclude_fields,
                                include_fields)

    mdvfile = mdv_common.MdvFile(prepare_for_read(filename))

    # value attributes
    az_deg, range_km, el_deg = mdvfile._calc_geometry()
    naz = len(az_deg)
    nele = len(el_deg)
    scan_type = mdvfile.projection

    if scan_type not in ['ppi', 'rhi']:
        raise NotImplementedError('No support for scan_type %s.' % scan_type)

    # time
    time = filemetadata('time')
    units = make_time_unit_str(mdvfile.times['time_begin'])
    time['units'] = units
    time_start = date2num(mdvfile.times['time_begin'], units)
    time_end = date2num(mdvfile.times['time_end'], units)
    time['data'] = np.linspace(time_start, time_end, naz * nele)

    # range
    _range = filemetadata('range')
    _range['data'] = np.array(range_km * 1000.0, dtype='float32')
    _range['meters_to_center_of_first_gate'] = _range['data'][0]
    _range['meters_between_gates'] = (_range['data'][1] - _range['data'][0])

    # fields
    fields = {}
    for mdv_field in set(mdvfile.fields):
        field_name = filemetadata.get_field_name(mdv_field)
        if field_name is None:
            continue

        # create and store the field dictionary
        field_dic = filemetadata(field_name)
        field_dic['_FillValue'] = get_fillvalue()
        dataextractor = mdv_common._MdvVolumeDataExtractor(
            mdvfile, mdvfile.fields.index(mdv_field), get_fillvalue())
        if delay_field_loading:
            field_dic = LazyLoadDict(field_dic)
            field_dic.set_lazy('data', dataextractor)
        else:
            field_dic['data'] = dataextractor()
        fields[field_name] = field_dic

    # metadata
    metadata = filemetadata('metadata')
    for meta_key, mdv_key in mdv_common.MDV_METADATA_MAP.items():
        metadata[meta_key] = mdvfile.master_header[mdv_key]

    # latitude
    latitude = filemetadata('latitude')
    latitude['data'] = np.array([mdvfile.radar_info['latitude_deg']],
                                dtype='float64')
    # longitude
    longitude = filemetadata('longitude')
    longitude['data'] = np.array([mdvfile.radar_info['longitude_deg']],
                                 dtype='float64')
    # altitude
    altitude = filemetadata('altitude')
    altitude['data'] = np.array([mdvfile.radar_info['altitude_km'] * 1000.0],
                                dtype='float64')

    # sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
    # sweep_end_ray_index
    sweep_number = filemetadata('sweep_number')
    sweep_mode = filemetadata('sweep_mode')
    fixed_angle = filemetadata('fixed_angle')
    sweep_start_ray_index = filemetadata('sweep_start_ray_index')
    sweep_end_ray_index = filemetadata('sweep_end_ray_index')
    len_time = len(time['data'])

    if scan_type == 'ppi':
        nsweeps = nele
        sweep_number['data'] = np.arange(nsweeps, dtype='int32')
        sweep_mode['data'] = np.array(
            nsweeps * ['azimuth_surveillance'], dtype='S')
        fixed_angle['data'] = np.array(el_deg, dtype='float32')
        sweep_start_ray_index['data'] = np.arange(0, len_time, naz,
                                                  dtype='int32')
        sweep_end_ray_index['data'] = np.arange(naz - 1, len_time, naz,
                                                dtype='int32')

    elif scan_type == 'rhi':
        nsweeps = naz
        sweep_number['data'] = np.arange(nsweeps, dtype='int32')
        sweep_mode['data'] = np.array(nsweeps * ['rhi'], dtype='S')
        fixed_angle['data'] = np.array(az_deg, dtype='float32')
        sweep_start_ray_index['data'] = np.arange(0, len_time, nele,
                                                  dtype='int32')
        sweep_end_ray_index['data'] = np.arange(nele - 1, len_time, nele,
                                                dtype='int32')

    # azimuth, elevation
    azimuth = filemetadata('azimuth')
    elevation = filemetadata('elevation')

    if scan_type == 'ppi':
        azimuth['data'] = np.tile(az_deg, nele)
        elevation['data'] = np.array(el_deg).repeat(naz)

    elif scan_type == 'rhi':
        azimuth['data'] = np.array(az_deg).repeat(nele)
        elevation['data'] = np.tile(el_deg, naz)

    # instrument parameters
    # we will set 4 parameters in the instrument_parameters dict
    # prt, prt_mode, unambiguous_range, and nyquist_velocity

    # TODO prt mode: Need to fix this.. assumes dual if two prts
    if mdvfile.radar_info['prt2_s'] == 0.0:
        prt_mode_str = 'fixed'
    else:
        prt_mode_str = 'dual'

    prt_mode = filemetadata('prt_mode')
    prt = filemetadata('prt')
    unambiguous_range = filemetadata('unambiguous_range')
    nyquist_velocity = filemetadata('nyquist_velocity')
    beam_width_h = filemetadata('radar_beam_width_h')
    beam_width_v = filemetadata('radar_beam_width_v')

    prt_mode['data'] = np.array([prt_mode_str] * nsweeps, dtype='S')
    prt['data'] = np.array([mdvfile.radar_info['prt_s']] * nele * naz,
                           dtype='float32')

    urange_m = mdvfile.radar_info['unambig_range_km'] * 1000.0
    unambiguous_range['data'] = np.array([urange_m] * naz * nele,
                                         dtype='float32')

    uvel_mps = mdvfile.radar_info['unambig_vel_mps']
    nyquist_velocity['data'] = np.array([uvel_mps] * naz * nele,
                                        dtype='float32')
    beam_width_h['data'] = np.array(
        [mdvfile.radar_info['horiz_beam_width_deg']], dtype='float32')
    beam_width_v['data'] = np.array(
        [mdvfile.radar_info['vert_beam_width_deg']], dtype='float32')

    instrument_parameters = {'prt_mode': prt_mode, 'prt': prt,
                             'unambiguous_range': unambiguous_range,
                             'nyquist_velocity': nyquist_velocity,
                             'radar_beam_width_h': beam_width_h,
                             'radar_beam_width_v': beam_width_v}

    if not delay_field_loading:
        mdvfile.close()
    return Radar(
        time, _range, fields, metadata, scan_type,
        latitude, longitude, altitude,
        sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth, elevation,
        instrument_parameters=instrument_parameters)



"""
pyart.correct.region_dealias
============================

Region based dealiasing using a dynamic network reduction for region joining.

.. autosummary::
    :toctree: generated/

    dealias_region_based
    _find_regions
    _find_sweep_interval_splits
    _combine_regions
    _edge_sum_and_count
    _cost_function
    _gradient

.. autosummary::
    :toctree: generated/
    :template: dev_template.rst

    _RegionTracker
    _EdgeTracker

"""

import warnings

import numpy as np
import scipy.ndimage as ndimage

from scipy.optimize import fmin_l_bfgs_b

from ..config import get_metadata, get_fillvalue
from ._common_dealias import _parse_fields, _parse_gatefilter, _set_limits
from ._common_dealias import _parse_rays_wrap_around, _parse_nyquist_vel
from ._fast_edge_finder import _fast_edge_finder


# Possible future improvements to the region based dealiasing algorithm:
#
# * Find connected regions in the entire volume (3D) an dealias the entire
#   volume as apposed to the current sweep-by-sweep approach.
# * Scale the Doppler velocities by the Nyquist interval so that
#   calculations are done in units of percentage of the Nyquist interval.
#   This scaling is already done in the _EdgeTracker object but could also be
#   applied before finding regions of similar velocities.
# * Currently no attmpt is made to unfold gates which are included in the
#   gatefilter.  One could add a second pass to the algorithms where gates
#   excluded by the filter are unfolded using a more elemenary approach.  In
#   this manner the filter could define only "good" gates which establish the
#   folding pattern which is then applied to all gates.
# * Improve perfornace by implementing a priority queue in the _EdgeTracker
#   object. See comments in the class for details.


def dealias_region_based(
        radar, ref_vel_field=None, interval_splits=3, interval_limits=None,
        skip_between_rays=100, skip_along_ray=100, centered=True,
        nyquist_vel=None, check_nyquist_uniform=True, gatefilter=False,
        rays_wrap_around=None, keep_original=False, set_limits=True,
        vel_field=None, corr_vel_field=None, **kwargs):
    """
    Dealias Doppler velocities using a region based algorithm.

    Performs Doppler velocity dealiasing by finding regions of similar
    velocities and unfolding and merging pairs of regions until all
    regions are unfolded. Unfolding and merging regions is accomplished by
    modeling the problem as a dynamic network reduction.

    Parameters
    ----------
    radar : Radar
        Radar object containing Doppler velocities to dealias.
    ref_vel_field : str or None, optional
         Field in radar containing a reference velocity field used to anchor
         the unfolded velocities once the algorithm completes. Typically this
         field is created by simulating the radial velocities from wind data
         from an atmospheric sonding using
         :py:func:`pyart.util.simulated_vel_from_profile`.
    interval_splits : int, optional
        Number of segments to split the nyquist interval into when finding
        regions of similar velocity. More splits creates a larger number of
        initial regions which takes longer to process but may result in better
        dealiasing. The default value of 3 seems to be a good compromise
        between performance and artifact free dealiasing. This value
        is not used if the interval_limits parameter is not None.
    interval_limits : array like or None, optional
        Velocity limits used for finding regions of similar velocity. Should
        cover the entire nyquist interval. None, the default value, will
        split the Nyquist interval into interval_splits equal sized
        intervals.
    skip_between_rays, skip_along_ray : int, optional
        Maximum number of filtered gates to skip over when joining regions,
        gaps between region larger than this will not be connected. Parameters
        specify the maximum number of filtered gates between and along a ray.
        Set these parameters to 0 to disable unfolding across filtered gates.
    centered : bool, optional
        True to apply centering to each sweep after the dealiasing algorithm
        so that the average number of unfolding is near 0. False does not
        apply centering which may results in individual sweeps under or over
        folded by the nyquist interval.
    nyquist_velocity : array like or float, optional
        Nyquist velocity in unit identical to those stored in the radar's
        velocity field, either for each sweep or a single value which will be
        used for all sweeps. None will attempt to determine this value from
        the Radar object.
    check_nyquist_uniform : bool, optional
        True to check if the Nyquist velocities are uniform for all rays
        within a sweep, False will skip this check. This parameter is ignored
        when the nyquist_velocity parameter is not None.
    gatefilter : GateFilter, None or False, optional.
        A GateFilter instance which specified which gates should be
        ignored when performing de-aliasing. A value of None created this
        filter from the radar moments using any additional arguments by
        passing them to :py:func:`moment_based_gate_filter`. False, the
        default, disables filtering including all gates in the dealiasing.
    rays_wrap_around : bool or None, optional
        True when the rays at the beginning of the sweep and end of the sweep
        should be interpreted as connected when de-aliasing (PPI scans).
        False if they edges should not be interpreted as connected (other scan
        types). None will determine the correct value from the radar
        scan type.
    keep_original : bool, optional
        True to retain the original Doppler velocity values at gates
        where the dealiasing procedure fails or was not applied. False
        does not replacement and these gates will be masked in the corrected
        velocity field.
    set_limits : bool, optional
        True to set valid_min and valid_max elements in the returned
        dictionary. False will not set these dictionary elements.
    vel_field : str, optional
        Field in radar to use as the Doppler velocities during dealiasing.
        None will use the default field name from the Py-ART configuration
        file.
    corr_vel_field : str, optional
        Name to use for the dealiased Doppler velocity field metadata. None
        will use the default field name from the Py-ART configuration file.

    Returns
    -------
    corr_vel : dict
        Field dictionary containing dealiased Doppler velocities. Dealiased
        array is stored under the 'data' key.

    """
    # parse function parameters
    vel_field, corr_vel_field = _parse_fields(vel_field, corr_vel_field)
    gatefilter = _parse_gatefilter(gatefilter, radar, **kwargs)
    rays_wrap_around = _parse_rays_wrap_around(rays_wrap_around, radar)
    nyquist_vel = _parse_nyquist_vel(nyquist_vel, radar, check_nyquist_uniform)

    # parse ref_vel_field
    if ref_vel_field is None:
        ref_vdata = None
    else:
        ref_vdata = radar.fields[ref_vel_field]['data']

    # exclude masked and invalid velocity gates
    gatefilter.exclude_masked(vel_field)
    gatefilter.exclude_invalid(vel_field)
    gfilter = gatefilter.gate_excluded

    # perform dealiasing
    vdata = radar.fields[vel_field]['data'].view(np.ndarray)
    data = vdata.copy()     # dealiased velocities

    # loop over sweeps
    for nsweep, sweep_slice in enumerate(radar.iter_slice()):

        # extract sweep data
        sdata = vdata[sweep_slice].copy()   # is a copy needed here?
        scorr = data[sweep_slice]
        sfilter = gfilter[sweep_slice]

        # find nyquist velocity and interval segmentation limits
        nyquist_interval = nyquist_vel[nsweep] * 2.
        if interval_limits is None:
            nvel = nyquist_vel[nsweep]
            valid_sdata = sdata[~sfilter]
            s_interval_limits = _find_sweep_interval_splits(
                nvel, interval_splits, valid_sdata, nsweep)
        else:
            s_interval_limits = interval_limits

        # find regions in original data
        labels, nfeatures = _find_regions(sdata, sfilter, s_interval_limits)
        # skip sweep if all gates are masked or only a single region
        if nfeatures < 2:
            continue
        bincount = np.bincount(labels.ravel())
        num_masked_gates = bincount[0]
        region_sizes = bincount[1:]

        # find all edges between regions
        indices, edge_count, velos = _edge_sum_and_count(
            labels, num_masked_gates, sdata, rays_wrap_around,
            skip_between_rays, skip_along_ray)

        # no unfolding required if no edges exist between regions
        if len(edge_count) == 0:
            continue

        # find the number of folds in the regions
        region_tracker = _RegionTracker(region_sizes)
        edge_tracker = _EdgeTracker(indices, edge_count, velos,
                                    nyquist_interval, nfeatures+1)
        while True:
            if _combine_regions(region_tracker, edge_tracker):
                break

        # center sweep if requested, determine a global sweep unfold number
        # so that the average number of gate folds is zero.
        if centered:
            gates_dealiased = region_sizes.sum()
            total_folds = np.sum(
                region_sizes * region_tracker.unwrap_number[1:])
            sweep_offset = int(round(float(total_folds) / gates_dealiased))
            if sweep_offset != 0:
                region_tracker.unwrap_number -= sweep_offset

        # dealias the data using the fold numbers
        nwrap = np.take(region_tracker.unwrap_number, labels)
        scorr += nwrap * nyquist_interval

        # anchor unfolded velocities against reference velocity
        if ref_vdata is not None:
            sref = ref_vdata[sweep_slice]
            gfold = (sref-scorr).mean()/nyquist_interval
            gfold = round(gfold)

            # Anchor specific regions against reference velocity
            # Do this by constraining cost function due to difference
            # from reference velocity and to 2D continuity
            new_interval_limits = np.linspace(scorr.min(), scorr.max(), 10)
            labels_corr, nfeatures_corr = _find_regions(
                scorr, sfilter, new_interval_limits)
            
            # If we only have one region, just adjust the whole sweep by
            # x nyquist intervals
            if nfeatures_corr < 2:
                scorr = scorr + gfold * nyquist_interval
            else:
                bounds_list = [(x, y) for (x, y) in zip(
                    -6*np.ones(nfeatures_corr), 5*np.ones(nfeatures_corr))]
                scorr_means = np.zeros(nfeatures_corr)
                sref_means = np.zeros(nfeatures_corr)
                for reg in range(1, nfeatures_corr+1):
                    scorr_means[reg-1] = np.ma.mean(scorr[labels_corr == reg])
                    sref_means[reg-1] = np.ma.mean(sref[labels_corr == reg])

                def cost_function(x):
                    return _cost_function(x, scorr_means, sref_means,
                                          nyquist_interval, nfeatures_corr)

                def gradient(x):
                    return _gradient(x, scorr_means, sref_means,
                                     nyquist_interval, nfeatures_corr)

                nyq_adjustments = fmin_l_bfgs_b(
                    cost_function, gfold*np.ones((nfeatures_corr)), disp=True,
                    fprime=gradient, bounds=bounds_list, maxiter=200,
                    )

                i = 0
                for reg in range(1, nfeatures_corr):
                    scorr[labels == reg] += (nyquist_interval *
                                             np.round(nyq_adjustments[0][i]))
                    i = i + 1

    # fill_value from the velocity dictionary if present
    fill_value = radar.fields[vel_field].get('_FillValue', get_fillvalue())

    # mask filtered gates
    if np.any(gfilter):
        data = np.ma.array(data, mask=gfilter, fill_value=fill_value)

    # restore original values where dealiasing not applied
    if keep_original:
        data[gfilter] = radar.fields[vel_field]['data'][gfilter]

    # return field dictionary containing dealiased Doppler velocities
    corr_vel = get_metadata(corr_vel_field)
    corr_vel['data'] = data
    corr_vel['_FillValue'] = fill_value

    if set_limits:
        # set valid_min and valid_max in corr_vel
        _set_limits(data, nyquist_vel, corr_vel)

    return corr_vel


def _find_sweep_interval_splits(nyquist, interval_splits, velocities, nsweep):
    """ Return the interval limits for a given sweep. """
    # The Nyquist interval is split into interval_splits  equal sized areas.
    # If velocities outside the Nyquist are present the number and
    # limits of the interval splits must be adjusted so that theses
    # velocities are included in one of the splits.
    add_start = add_end = 0
    interval = (2. * nyquist) / (interval_splits)
    # no change from default if all gates filtered
    if len(velocities) != 0:
        max_vel = velocities.max()
        min_vel = velocities.min()
        if max_vel > nyquist or min_vel < -nyquist:
            msg = ("Velocities outside of the Nyquist interval found in "
                   "sweep %i." % (nsweep))
            warnings.warn(msg, UserWarning)
            # additional intervals must be added to capture the velocities
            # outside the nyquist limits
            add_start = int(np.ceil((max_vel - nyquist) / (interval)))
            add_end = int(np.ceil(-(min_vel + nyquist) / (interval)))

    start = -nyquist - add_start * interval
    end = nyquist + add_end * interval
    num = interval_splits + 1 + add_start + add_end
    return np.linspace(start, end, num, endpoint=True)


def _find_regions(vel, gfilter, limits):
    """
    Find regions of similar velocity.

    For each pair of values in the limits array (or list) find all connected
    velocity regions within these limits.

    Parameters
    ----------
    vel : 2D ndarray
        Array containing velocity data for a single sweep.
    gfilter : 2D ndarray
        Filter indicating if a particular gate should be masked. True
        indicates the gate should be masked (excluded).
    limits : array like
        Velocity limits for region finding. For each pair of limits, taken
        from elements i and i+1 of the array, all connected regions with
        velocities within these limits will be found.

    Returns
    -------
    label : ndarray
        Interger array with each region labeled by a value. The array
        ranges from 0 to nfeatures, inclusive, where a value of 0 indicates
        masked gates and non-zero indicates a region of connected gates.
    nfeatures : int
        Number of regions found.

    """
    mask = ~gfilter
    label = np.zeros(vel.shape, dtype=np.int32)
    nfeatures = 0
    for lmin, lmax in zip(limits[:-1], limits[1:]):

        # find connected regions within the limits
        inp = (lmin <= vel) & (vel < lmax) & mask
        limit_label, limit_nfeatures = ndimage.label(inp)

        # add these regions to the global regions
        limit_label[np.nonzero(limit_label)] += nfeatures
        label += limit_label
        nfeatures += limit_nfeatures

    return label, nfeatures


def _edge_sum_and_count(labels, num_masked_gates, data,
                        rays_wrap_around, max_gap_x, max_gap_y):
    """
    Find all edges between labels regions.

    Returns the indices, count and velocities of all edges.
    """
    total_nodes = labels.shape[0] * labels.shape[1] - num_masked_gates
    if rays_wrap_around:
        total_nodes += labels.shape[0] * 2

    indices, velocities = _fast_edge_finder(
        labels.astype('int32'), data.astype('float32'),
        rays_wrap_around, max_gap_x, max_gap_y, total_nodes)
    index1, index2 = indices
    vel1, vel2 = velocities
    count = np.ones_like(vel1, dtype=np.int32)

    # return early if not edges were found
    if len(vel1) == 0:
        return ([], []), [], ([], [])

    # find the unique edges, procedure based on method in
    # scipy.sparse.coo_matrix.sum_duplicates
    # except we have three data arrays, vel1, vel2, and count
    order = np.lexsort((index1, index2))
    index1 = index1[order]
    index2 = index2[order]
    vel1 = vel1[order]
    vel2 = vel2[order]
    count = count[order]

    unique_mask = ((index1[1:] != index1[:-1]) |
                   (index2[1:] != index2[:-1]))
    unique_mask = np.append(True, unique_mask)
    index1 = index1[unique_mask]
    index2 = index2[unique_mask]

    unique_inds, = np.nonzero(unique_mask)
    vel1 = np.add.reduceat(vel1, unique_inds, dtype=vel1.dtype)
    vel2 = np.add.reduceat(vel2, unique_inds, dtype=vel2.dtype)
    count = np.add.reduceat(count, unique_inds, dtype=count.dtype)

    return (index1, index2), count, (vel1, vel2)


def _combine_regions(region_tracker, edge_tracker):
    """ Returns True when done. """
    # Edge parameters from edge with largest weight
    status, extra = edge_tracker.pop_edge()
    if status:
        return True
    node1, node2, weight, diff, edge_number = extra
    rdiff = int(np.round(diff))

    # node sizes of nodes to be merged
    node1_size = region_tracker.get_node_size(node1)
    node2_size = region_tracker.get_node_size(node2)

    # determine which nodes should be merged
    if node1_size > node2_size:
        base_node, merge_node = node1, node2
    else:
        base_node, merge_node = node2, node1
        rdiff = -rdiff

    # unwrap merge_node
    if rdiff != 0:
        region_tracker.unwrap_node(merge_node, rdiff)
        edge_tracker.unwrap_node(merge_node, rdiff)

    # merge nodes
    region_tracker.merge_nodes(base_node, merge_node)
    edge_tracker.merge_nodes(base_node, merge_node, edge_number)

    return False


# Minimize cost function that is sum of difference between regions and
# sounding
def _cost_function(nyq_vector, vels_slice_means,
                   svels_slice_means, v_nyq_vel, nfeatures):
    """ Cost function for minimization in region based algorithm. """
    cost = 0
    i = 0

    for reg in range(nfeatures):
        add_value = 0
        # Deviance from sounding
        add_value = 1*(vels_slice_means[reg] +
                       np.round(nyq_vector[i])*v_nyq_vel -
                       svels_slice_means[reg])**2

        # Region continuity
        vels_without_cur = np.delete(vels_slice_means, reg)
        diffs = np.square(vels_slice_means[reg]-vels_without_cur)
        if len(diffs) > 0:
            the_min = np.argmin(diffs)
            vel_wo_cur = vels_without_cur[the_min]
        else:
            vel_wo_cur = vels_slice_means[reg]
        add_value2 = np.square(vels_slice_means[reg] +
                               np.round(nyq_vector[i])*v_nyq_vel -
                               vel_wo_cur)
        if np.isfinite(add_value2):
            add_value += .1*add_value2

        if np.isfinite(add_value):
            cost += add_value
        i = i + 1

    return cost


def _gradient(nyq_vector, vels_slice_means, svels_slice_means,
              v_nyq_vel, nfeatures):
    """ Gradient of cost function for minimization
        in region based algorithm. """
    gradient_vector = np.zeros(len(nyq_vector))
    i = 0
    for reg in range(nfeatures):
        add_value = (vels_slice_means[reg] +
                     np.round(nyq_vector[i])*v_nyq_vel -
                     svels_slice_means[reg])
        if np.isfinite(add_value):
            gradient_vector[i] = 2*add_value*v_nyq_vel

        # Regional continuity
        vels_without_cur = np.delete(vels_slice_means, reg)
        diffs = np.square(vels_slice_means[reg]-vels_without_cur)
        if len(diffs) > 0:
            the_min = np.argmin(diffs)
            vel_wo_cur = vels_without_cur[the_min]
        else:
            vel_wo_cur = vels_slice_means[reg]

        add_value2 = (vels_slice_means[reg] +
                      np.round(nyq_vector[i])*v_nyq_vel -
                      vel_wo_cur)

        if np.isfinite(add_value2):
            gradient_vector[i] += 2*.1*add_value2*v_nyq_vel

        i = i + 1

    return gradient_vector


class _RegionTracker(object):
    """
    Tracks the location of radar volume regions contained in each node
    as the network is reduced.
    """

    def __init__(self, region_sizes):
        """ initalize. """
        # number of gates in each node
        nregions = len(region_sizes) + 1
        self.node_size = np.zeros(nregions, dtype='int32')
        self.node_size[1:] = region_sizes[:]

        # array of lists containing the regions in each node
        self.regions_in_node = np.zeros(nregions, dtype='object')
        for i in range(nregions):
            self.regions_in_node[i] = [i]

        # number of unwrappings to apply to dealias each region
        self.unwrap_number = np.zeros(nregions, dtype='int32')

    def merge_nodes(self, node_a, node_b):
        """ Merge node b into node a. """

        # move all regions from node_b to node_a
        regions_to_merge = self.regions_in_node[node_b]
        self.regions_in_node[node_a].extend(regions_to_merge)
        self.regions_in_node[node_b] = []

        # update node sizes
        self.node_size[node_a] += self.node_size[node_b]
        self.node_size[node_b] = 0
        return

    def unwrap_node(self, node, nwrap):
        """ Unwrap all gates contained a node. """
        if nwrap == 0:
            return
        # for each region in node add nwrap
        regions_to_unwrap = self.regions_in_node[node]
        self.unwrap_number[regions_to_unwrap] += nwrap
        return

    def get_node_size(self, node):
        """ Return the number of gates in a node. """
        return self.node_size[node]


class _EdgeTracker(object):
    """ A class for tracking edges in a dynamic network. """

    def __init__(self, indices, edge_count, velocities, nyquist_interval,
                 nnodes):
        """ initialize """

        nedges = int(len(indices[0]) / 2)

        # node number and different in sum for each edge
        self.node_alpha = np.zeros(nedges, dtype=np.int32)
        self.node_beta = np.zeros(nedges, dtype=np.int32)
        self.sum_diff = np.zeros(nedges, dtype=np.float32)

        # number of connections between the regions
        self.weight = np.zeros(nedges, dtype=np.int32)

        # fast finding
        self._common_finder = np.zeros(nnodes, dtype=np.bool)
        self._common_index = np.zeros(nnodes, dtype=np.int32)
        self._last_base_node = -1

        # array of linked lists pointing to each node
        self.edges_in_node = np.zeros(nnodes, dtype='object')
        for i in range(nnodes):
            self.edges_in_node[i] = []

        # fill out data from the provides indicies, edge counts and velocities
        edge = 0
        idx1, idx2 = indices
        vel1, vel2 = velocities
        for i, j, count, vel, nvel in zip(idx1, idx2, edge_count, vel1, vel2):
            if i < j:
                continue
            self.node_alpha[edge] = i
            self.node_beta[edge] = j
            self.sum_diff[edge] = ((vel - nvel) / nyquist_interval)
            self.weight[edge] = count
            self.edges_in_node[i].append(edge)
            self.edges_in_node[j].append(edge)

            edge += 1

        # list which orders edges according to their weight, highest first
        # TODO
        self.priority_queue = []

    def merge_nodes(self, base_node, merge_node, foo_edge):
        """ Merge nodes. """

        # remove edge between base and merge nodes
        self.weight[foo_edge] = -999
        self.edges_in_node[merge_node].remove(foo_edge)
        self.edges_in_node[base_node].remove(foo_edge)
        self._common_finder[merge_node] = False

        # find all the edges in the two nodes
        edges_in_merge = list(self.edges_in_node[merge_node])

        # loop over base_node edges if last base_node was different
        if self._last_base_node != base_node:
            self._common_finder[:] = False
            edges_in_base = list(self.edges_in_node[base_node])
            for edge_num in edges_in_base:

                # reverse edge if needed so node_alpha is base_node
                if self.node_beta[edge_num] == base_node:
                    self._reverse_edge_direction(edge_num)
                assert self.node_alpha[edge_num] == base_node

                # find all neighboring nodes to base_node
                neighbor = self.node_beta[edge_num]
                self._common_finder[neighbor] = True
                self._common_index[neighbor] = edge_num

        # loop over edge nodes
        for edge_num in edges_in_merge:

            # reverse edge so that node alpha is the merge_node
            if self.node_beta[edge_num] == merge_node:
                self._reverse_edge_direction(edge_num)
            assert self.node_alpha[edge_num] == merge_node

            # update all the edges to point to the base node
            self.node_alpha[edge_num] = base_node

            # if base_node also has an edge with the neighbor combine them
            neighbor = self.node_beta[edge_num]
            if self._common_finder[neighbor]:
                base_edge_num = self._common_index[neighbor]
                self._combine_edges(base_edge_num, edge_num,
                                    merge_node, neighbor)
            # if not fill in _common_ arrays.
            else:
                self._common_finder[neighbor] = True
                self._common_index[neighbor] = edge_num

        # move all edges from merge_node to base_node
        edges = self.edges_in_node[merge_node]
        self.edges_in_node[base_node].extend(edges)
        self.edges_in_node[merge_node] = []
        self._last_base_node = int(base_node)
        return

    def _combine_edges(self, base_edge, merge_edge,
                       merge_node, neighbor_node):
        """ Combine edges into a single edge.  """
        # Merging nodes MUST be set to alpha prior to calling this function

        # combine edge weights
        self.weight[base_edge] += self.weight[merge_edge]
        self.weight[merge_edge] = -999.

        # combine sums
        self.sum_diff[base_edge] += self.sum_diff[merge_edge]

        # remove merge_edge from both node lists
        self.edges_in_node[merge_node].remove(merge_edge)
        self.edges_in_node[neighbor_node].remove(merge_edge)

        # TODO priority queue
        # remove merge_edge from edge priority queue
        # self.priority_queue.remove(merge_edge)

        # relocate base_edge in priority queue
        # self.priority_queue.sort()

    def _reverse_edge_direction(self, edge):
        """ Reverse an edges direction, change alpha and beta. """
        # swap nodes
        old_alpha = int(self.node_alpha[edge])
        old_beta = int(self.node_beta[edge])
        self.node_alpha[edge] = old_beta
        self.node_beta[edge] = old_alpha
        # swap sums
        self.sum_diff[edge] = -1. * self.sum_diff[edge]
        return

    def unwrap_node(self, node, nwrap):
        """ Unwrap a node. """
        if nwrap == 0:
            return
        # add weight * nwrap to each edge in node
        for edge in self.edges_in_node[node]:
            weight = self.weight[edge]
            if node == self.node_alpha[edge]:
                self.sum_diff[edge] += weight * nwrap
            else:
                assert self.node_beta[edge] == node
                self.sum_diff[edge] += -weight * nwrap
        return

    def pop_edge(self):
        """ Pop edge with largest weight. Return node numbers and diff. """

        # if len(priority_queue) == 0:
        #     return True, None

        # edge_num = self.priority_queue[0]
        edge_num = np.argmax(self.weight)
        node1 = self.node_alpha[edge_num]
        node2 = self.node_beta[edge_num]
        weight = self.weight[edge_num]
        diff = self.sum_diff[edge_num] / (float(weight))

        if weight < 0:
            return True, None
        return False, (node1, node2, weight, diff, edge_num)



"""
====================================================================
Create an RHI plot with reflectivity contour lines from an MDV file
====================================================================

An example which creates an RHI plot of reflectivity using a RadarDisplay object
and adding differnential Reflectivity contours from the same MDV file.

"""
print(__doc__)

# Author: Cory Weber (cweber@anl.gov)
# License: BSD 3 clause
import matplotlib.pyplot as plt
import pyart
import numpy as np
import scipy.ndimage as ndimage

filename = '220629.mdv'

# create the plot using RadarDisplay
sweep = 0
# read file
radar = pyart.io.read_mdv(filename)
display = pyart.graph.RadarDisplay(radar)
fig = plt.figure(figsize=[20, 5])
ax = fig.add_subplot(111)

# plot reflectivity
# alpha=0.25 sets the transparency of the pcolormesh to 75% transparent against
# the default white. matplolib overlaps the edges of the pcolormesh and creates
# a visable border around the edge, even with the default of edgecolor set to
# 'none' the transparancy is effected. the flowing paramters are designed to
# compensate for that:
# edgecolors=(1.0, 1.0, 1.0, 0.1) sets the lines between patches to nearly
# transparent
# linewidth=0.00015 makes lines between patches very small
# antialiased=true removes moire patterns.

display.plot('reflectivity', sweep=sweep, vmin=-8, vmax=64.0, fig=fig,
             ax=ax, colorbar_label='Reflectivity (dB)', alpha=0.75,
             edgecolors=(0.5, 0.5, 0.5, 0.3), linewidth=0.001,
             antialiased=True)

# Normal no alpha
# display.plot('reflectivity', sweep=sweep, vmin=-8, vmax=64.0, fig=fig,
#             ax=ax, colorbar_label='Reflectivity (dB)', antialiased=True)

# get data
start = radar.get_start(sweep)
end = radar.get_end(sweep) + 1
data = radar.get_field(sweep, 'differential_reflectivity')
x, y, z = radar.get_gate_x_y_z(sweep, edges=False)

x /= 1000.0
y /= 1000.0
z /= 1000.0

# apply a gaussian blur to the data set for nice smooth lines:
# sigma adjusts the distance effect of blending each cell,
# 4 is arbirarly set for visual impact.
data = ndimage.gaussian_filter(data, sigma=4)

# calculate (R)ange
R = np.sqrt(x ** 2 + y ** 2) * np.sign(y)
R = -R
display.set_limits(xlim=[0, 40], ylim=[0, 15])

# add contours
# creates steps 35 to 100 by 5
levels = np.arange(-3, 4, 0.25)
# levels_rain = np.arange(1, 4, 0.5)
levels_ice = np.arange(-2, -0, 0.5)
levels_rain = [0.75]

# adds contours to plot
contours = ax.contour(R, z, data, levels, linewidths=1, colors='k',
                      antialiased=True)
# adds more contours for ice and rain, matplotlib supports multiple sets of
# contours ice
contours_ice = ax.contour(R, z, data, levels_ice, linewidths=2, colors='blue')
# contours heavy rain
contours_rain = ax.contour(R, z, data, levels_rain, linewidths=2, colors='red')

# adds contour labels (fmt= '%r' displays 10.0 vs 10.0000)
plt.clabel(contours, levels, fmt='%r', inline=True, fontsize=10)

plt.show()



"""
====================================
Create a RHI plot from a Sigmet file
====================================

An example which creates a RHI plot of a Sigmet file using a RadarDisplay
object.

"""
print(__doc__)

# Author: Jonathan J. Helmus (jhelmus@anl.gov)
# License: BSD 3 clause

import matplotlib.pyplot as plt
import pyart
import netCDF4

filename = 'XSW110520113537.RAW7HHL'

# create the plot using RadarDisplay
radar = pyart.io.read_rsl(filename)
display = pyart.graph.RadarDisplay(radar)

fig = plt.figure(figsize=[10, 4])
ax = fig.add_subplot(111)

instrument_name = radar.metadata['instrument_name'].decode('utf-8')
time_start = netCDF4.num2date(radar.time['data'][0], radar.time['units'])
time_text = ' ' + time_start.isoformat() + 'Z '
azimuth = radar.fixed_angle['data'][0]
title = 'RHI ' + instrument_name + time_text + 'Azimuth %.2f' % (azimuth)

display.plot('reflectivity', 0, vmin=-32, vmax=64,
             title=title, colorbar_flag=False, ax=ax)
display.set_limits(ylim=[0, 17])

cax = fig.add_axes([.9, .1, 0.02, .8])
colorbar_label = 'Eq refl fact (dBz)'
display.plot_colorbar(fig=fig, cax=cax, label=colorbar_label)

plt.show()



"""
pyart.aux_io.rainbow_psr
========================

Routines for reading RAINBOW PSR files (Used by SELEX)

.. autosummary::
    :toctree: generated/

    read_rainbow_psr
    read_rainbow_psr_spectra
    read_psr_header
    read_psr_cpi_header
    read_psr_spectra
    get_item_numbers
    get_field
    get_spectra_field
    get_Doppler_info
    get_noise_field
    convert_data
    get_library
    get_library_path

"""

# specific modules for this function
import os
from warnings import warn
import ctypes
from copy import deepcopy

import numpy as np

from ..config import FileMetadata
from ..io.common import _test_arguments
from ..core.radar_spectra import RadarSpectra
from ..exceptions import MissingOptionalDependency
from ..util import cut_radar, ma_broadcast_to
from .rainbow_wrl import read_rainbow_wrl

# Check existence of required libraries
try:
    # `read_rainbow` as of wradlib version 1.0.0
    from wradlib.io import read_Rainbow as read_rainbow
    _WRADLIB_AVAILABLE = True
except ImportError:
    try:
        from wradlib.io import read_rainbow
        _WRADLIB_AVAILABLE = True
    except ImportError:
        _WRADLIB_AVAILABLE = False

PSR_FIELD_NAMES = {
    'TXh': 'transmitted_power_h',
    'TXv': 'transmitted_power_v',
    'NADUhh': 'spectral_noise_power_hh_ADU',
    'NADUvv': 'spectral_noise_power_vv_ADU',
    'NADUhv': 'spectral_noise_power_hv_ADU',
    'NADUvh': 'spectral_noise_power_vh_ADU',
    'ShhADU': 'complex_spectra_hh_ADU',
    'SvvADU': 'complex_spectra_vv_ADU',
}


def read_rainbow_psr(filename, filenames_psr, field_names=None,
                     additional_metadata=None, file_field_names=False,
                     exclude_fields=None, include_fields=None,
                     undo_txcorr=True, cpi='mean', ang_tol=0.5, azi_min=None,
                     azi_max=None, ele_min=None, ele_max=None, rng_min=None,
                     rng_max=None, **kwargs):
    """
    Read a PSR file.

    Parameters
    ----------
    filename : str
        Name of the rainbow file to be used as reference.
    filenames_psr : list of str
        Name of the PSR files
    field_names : dict, optional
        Dictionary mapping RAINBOW field names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        Py-ART configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    file_field_names : bool, optional
        True to use the MDV data type names for the field names. If this
        case the field_names parameter is ignored. The field dictionary will
        likely only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.
    undo_txcorr: Bool
        If True the correction of the transmitted power is removed from the
        noise signal
    cpi : str
        The CPI to use. Can be 'low_prf', 'intermediate_prf', 'high_prf',
        'mean', 'all'. If 'mean' the mean within the angle step is taken
    ang_tol : float
        Tolerated angle distance between nominal radar angle and angle in
        PSR files
    azi_min, azi_max, ele_min, ele_max : float or None
        The minimum and maximum angles to keep (deg)
    rng_min, rng_max : float or None
        The minimum and maximum ranges to keep (m)

    Returns
    -------
    radar : Radar
        Radar object containing data from PSR file.

    """
    # check that wradlib is available
    if not _WRADLIB_AVAILABLE:
        raise MissingOptionalDependency(
            "wradlib is required to use read_rainbow_psr but is not installed")

    # test for non empty kwargs
    _test_arguments(kwargs)

    # create radar object used as reference
    try:
        radar = read_rainbow_wrl(filename)
        radar.fields = dict()
        radar = cut_radar(
            radar, None, rng_min=rng_min, rng_max=rng_max, ele_min=ele_min,
            ele_max=ele_max, azi_min=azi_min, azi_max=azi_max)
    except OSError as ee:
        warn(str(ee))
        warn('Unable to read file '+filename)
        return None

    # create metadata retrieval object
    if field_names is None:
        field_names = PSR_FIELD_NAMES
    filemetadata = FileMetadata('PSR', field_names, additional_metadata,
                                file_field_names, exclude_fields,
                                include_fields)
    dBADU_to_dBm_hh = filemetadata('dBADU_to_dBm_hh')
    dBADU_to_dBm_vv = filemetadata('dBADU_to_dBm_vv')
    mfloss_h = filemetadata('matched_filter_loss_h')
    mfloss_v = filemetadata('matched_filter_loss_v')
    pathatt = filemetadata('path_attenuation')

    cpi_header, header = read_psr_cpi_headers(filenames_psr)

    if cpi_header is None:
        return None

    # keep only valid items
    prfs = np.sort(np.unique(cpi_header['prfs']))
    items, radar = get_item_numbers(
        radar, cpi_header['azi_start'], cpi_header['azi_stop'],
        cpi_header['ele_start'], cpi_header['ele_stop'], cpi_header['prfs'],
        prfs, cpi=cpi, ang_tol=ang_tol)

    if items.size == 0:
        warn('No items matching radar object')
        return None

    for field_name in field_names:
        field_data = get_field(
            radar, cpi_header, header, items, prfs.size, field_name,
            undo_txcorr=undo_txcorr, cpi=cpi)

        if field_name in ('noisedBADU_hh', 'noisedBADU_vv', 'noisedBm_hh',
                          'noisedBm_vv', 'noisedBZ_hh', 'noisedBZ_vv'):
            field_data = get_noise_field(
                radar, field_data, header, field_name)

        field_dict = filemetadata(field_name)
        field_dict['data'] = field_data
        radar.add_field(field_name, field_dict)

    # get further metadata
    pw_ind = header['states.spbpwidth']

    dBADU_to_dBm_hh['data'] = np.array([
        header['states.spbdbmtologoffset'][pw_ind]])
    radar.radar_calibration.update({'dBADU_to_dBm_hh': dBADU_to_dBm_hh})

    dBADU_to_dBm_vv['data'] = np.array([
        header['states.spbdpvdbmtologoffset'][pw_ind]])
    radar.radar_calibration.update({'dBADU_to_dBm_vv': dBADU_to_dBm_vv})

    mfloss_h['data'] = np.array([header['states.gdrxmfloss'][pw_ind]])
    mfloss_v['data'] = np.array([header['states.gdrxmfloss'][pw_ind]])
    radar.radar_calibration.update({'matched_filter_loss_h': mfloss_h})
    radar.radar_calibration.update({'matched_filter_loss_v': mfloss_v})

    pathatt['data'] = np.array([header['states.rspathatt']])
    radar.radar_calibration.update({'path_attenuation': pathatt})

    return radar


def read_rainbow_psr_spectra(filename, filenames_psr, field_names=None,
                             additional_metadata=None, file_field_names=False,
                             exclude_fields=None, include_fields=None,
                             undo_txcorr=True, fold=True, positive_away=True,
                             cpi='low_prf', ang_tol=0.5, azi_min=None,
                             azi_max=None, ele_min=None, ele_max=None,
                             rng_min=None, rng_max=None, **kwargs):
    """
    Read a PSR file to get the complex spectra

    Parameters
    ----------
    filename : str
        Name of the rainbow file to be used as reference.
    filenames_psr : list of str
        list of PSR file names
    field_names : dict, optional
        Dictionary mapping RAINBOW field names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        Py-ART configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    file_field_names : bool, optional
        True to use the MDV data type names for the field names. If this
        case the field_names parameter is ignored. The field dictionary will
        likely only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.
    undo_txcorr: Bool
        If True the correction of the transmitted power is removed from the
        noise signal
    fold: Bool
        If True the spectra is folded so that 0-Doppler is in the middle
    positive_away: Bool
        If True the spectra is reversed so that positive velocities are
        away from the radar
    cpi : str
        The CPI to use. Can be 'low_prf', 'intermediate_prf', 'high_prf' or
        'all'
    ang_tol : float
        Tolerated angle distance between nominal radar angle and angle in
        PSR files
    azi_min, azi_max, ele_min, ele_max : float or None
        The minimum and maximum angles to keep (deg)
    rng_min, rng_max : float or None
        The minimum and maximum ranges to keep (m)

    Returns
    -------
    radar : Radar
        Radar object containing data from PSR file.

    """
    # check that wradlib is available
    if not _WRADLIB_AVAILABLE:
        raise MissingOptionalDependency(
            "wradlib is required to use read_rainbow_psr but is not installed")

    # test for non empty kwargs
    _test_arguments(kwargs)

    # create radar object used as reference
    try:
        radar = read_rainbow_wrl(filename)
        radar.fields = dict()
        rng_orig = radar.range['data']
        radar = cut_radar(
            radar, None, rng_min=rng_min, rng_max=rng_max, ele_min=ele_min,
            ele_max=ele_max, azi_min=azi_min, azi_max=azi_max)
        if radar is None:
            warn('No data within specified azimuth, elevation and'
                 ' range limits')
            return None
        ind_rng_start = np.where(rng_orig == radar.range['data'][0])[0]
        ind_rng_end = np.where(rng_orig == radar.range['data'][-1])[0]
        ind_rng = np.arange(ind_rng_start, ind_rng_end+1, dtype=int)
    except OSError as ee:
        warn(str(ee))
        warn('Unable to read file '+filename)
        return None

    # create metadata retrieval object
    if field_names is None:
        field_names = PSR_FIELD_NAMES
    filemetadata = FileMetadata('PSR', field_names, additional_metadata,
                                file_field_names, exclude_fields,
                                include_fields)
    dBADU_to_dBm_hh = filemetadata('dBADU_to_dBm_hh')
    dBADU_to_dBm_vv = filemetadata('dBADU_to_dBm_vv')
    mfloss_h = filemetadata('matched_filter_loss_h')
    mfloss_v = filemetadata('matched_filter_loss_v')
    pathatt = filemetadata('path_attenuation')

    cpi_header, header = read_psr_cpi_headers(filenames_psr)

    if cpi_header is None:
        return None

    # keep only valid items
    prfs = np.sort(np.unique(cpi_header['prfs']))
    items, radar = get_item_numbers(
        radar, cpi_header['azi_start'], cpi_header['azi_stop'],
        cpi_header['ele_start'], cpi_header['ele_stop'], cpi_header['prfs'],
        prfs, cpi=cpi, ang_tol=ang_tol)

    if items.size == 0:
        warn('No items matching radar object')
        return None

    fields = {}
    for field_name in field_names:
        if field_name in ('spectral_noise_power_hh_ADU',
                          'spectral_noise_power_vv_ADU',
                          'spectral_noise_power_hv_ADU',
                          'spectral_noise_power_vh_ADU'):
            field_data = get_spectral_noise(
                radar, cpi_header, header, items, undo_txcorr=undo_txcorr)
        else:
            field_data = get_spectra_field(
                radar, filenames_psr, cpi_header['npulses'],
                header['items_per_file'], items, ind_rng, fold=fold,
                positive_away=positive_away)

        field_dict = filemetadata(field_name)
        field_dict['data'] = field_data
        fields[field_name] = field_dict

    Doppler_velocity = filemetadata('Doppler_velocity')
    Doppler_frequency = filemetadata('Doppler_frequency')
    npulses = filemetadata('number_of_pulses')

    vel_data, freq_data = get_Doppler_info(
        cpi_header['prfs'][items], cpi_header['npulses'][items],
        header['states.rsplambda']*1e-2, fold=fold)

    Doppler_velocity['data'] = vel_data
    Doppler_frequency['data'] = freq_data
    npulses['data'] = cpi_header['npulses'][items]

    # get further metadata
    pw_ind = header['states.spbpwidth']

    dBADU_to_dBm_hh['data'] = np.array([
        header['states.spbdbmtologoffset'][pw_ind]])
    radar.radar_calibration.update({'dBADU_to_dBm_hh': dBADU_to_dBm_hh})

    dBADU_to_dBm_vv['data'] = np.array([
        header['states.spbdpvdbmtologoffset'][pw_ind]])
    radar.radar_calibration.update({'dBADU_to_dBm_vv': dBADU_to_dBm_vv})

    mfloss_h['data'] = np.array([header['states.gdrxmfloss'][pw_ind]])
    mfloss_v['data'] = np.array([header['states.gdrxmfloss'][pw_ind]])
    radar.radar_calibration.update({'matched_filter_loss_h': mfloss_h})
    radar.radar_calibration.update({'matched_filter_loss_v': mfloss_v})

    pathatt['data'] = np.array([header['states.rspathatt']])
    radar.radar_calibration.update({'path_attenuation': pathatt})

    return RadarSpectra(
        radar.time, radar.range, fields, radar.metadata,
        radar.scan_type, radar.latitude, radar.longitude, radar.altitude,
        radar.sweep_number, radar.sweep_mode, radar.fixed_angle,
        radar.sweep_start_ray_index, radar.sweep_end_ray_index,
        radar.azimuth, radar.elevation, npulses,
        Doppler_velocity=Doppler_velocity, Doppler_frequency=Doppler_frequency,
        rays_are_indexed=radar.rays_are_indexed,
        ray_angle_res=radar.ray_angle_res,
        instrument_parameters=radar.instrument_parameters,
        radar_calibration=radar.radar_calibration)


def read_psr_cpi_headers(filenames):
    """
    Reads the CPI data headers contained in multiple PSR files

    Parameters
    ----------
    filenames : list of str
        Name of the PSR files

    Returns
    -------
    cpi_header, header : dict
        Dictionary containing the PSR header data and the CPI headers data

    """
    cpi_header = {
        'azi_start': [],
        'azi_stop': [],
        'ele_start': [],
        'ele_stop': [],
        'npulses': [],
        'prfs': [],
        'ngates': [],
        'tx_pwr': [],
        'noise': [],
    }

    header = None
    for filename in filenames:
        cpi_header_aux, header_aux = read_psr_cpi_header(filename)
        if cpi_header_aux is None:
            warn('File '+filename+' could not be read')
            continue

        if header is None:
            header = header_aux
            header.update({'items_per_file': [header_aux['item.count']]})
        else:
            header['item.count'] = (
                header['item.count']+header_aux['item.count'])
            header['items_per_file'].append(header_aux['item.count'])

        cpi_header['azi_start'].extend(cpi_header_aux['azi_start'])
        cpi_header['azi_stop'].extend(cpi_header_aux['azi_stop'])
        cpi_header['ele_start'].extend(cpi_header_aux['ele_start'])
        cpi_header['ele_stop'].extend(cpi_header_aux['ele_stop'])
        cpi_header['npulses'].extend(cpi_header_aux['npulses'])
        cpi_header['prfs'].extend(cpi_header_aux['prfs'])
        cpi_header['ngates'].extend(cpi_header_aux['ngates'])
        cpi_header['tx_pwr'].extend(cpi_header_aux['tx_pwr'])

        if 'noise' not in cpi_header_aux:
            cpi_header['noise'] = None

        if cpi_header['noise'] is not None:
            cpi_header['noise'].extend(cpi_header_aux['noise'])

    cpi_header['azi_start'] = np.array(cpi_header['azi_start'])
    cpi_header['azi_stop'] = np.array(cpi_header['azi_stop'])
    cpi_header['ele_start'] = np.array(cpi_header['ele_start'])
    cpi_header['ele_stop'] = np.array(cpi_header['ele_stop'])
    cpi_header['npulses'] = np.array(cpi_header['npulses'])
    cpi_header['prfs'] = np.array(cpi_header['prfs'])
    cpi_header['ngates'] = np.array(cpi_header['ngates'])
    cpi_header['tx_pwr'] = np.ma.array(cpi_header['tx_pwr'])
    cpi_header['noise'] = np.ma.array(cpi_header['noise'])
    header['items_per_file'] = np.array(header['items_per_file'])

    return cpi_header, header


def read_psr_header(filename):
    """
    Read a PSR file header.

    Parameters
    ----------
    filename : str
        Name of the PSR file

    Returns
    -------
    header : dict
        Dictionary containing the PSR header data

    """
    header = dict()
    try:
        with open(filename, 'r', newline=None, encoding='latin-1') as txtfile:
            # read first line: ARCHIVE_HEADER_START
            line = txtfile.readline()
            while 0 == 0:
                line = txtfile.readline()

                if 'ARCHIVE_HEADER_END' in line:
                    break

                strings = line.split('=')
                key = strings[0]
                val = convert_data(strings[1].replace('\n', ''))
                header.update({key: val})

            # read noise
            line = txtfile.readline()
            if 'PSR REDUCED' not in line:
                warn('Header does not contain noise data')
                return header

            noise = np.array([])
            while 0 == 0:
                line = txtfile.readline()

                if not line.startswith('i'):
                    break

                strings = line.split()
                noise = np.append(noise, np.float32(strings[3].split('=')[1]))

            header.update({'noise': noise})

            return header
    except EnvironmentError as ee:
        warn(str(ee))
        warn('Unable to read file '+filename)
        return header


def read_psr_cpi_header(filename):
    """
    Reads the CPI data headers contained in a PSR file

    Parameters
    ----------
    filename : str
        Name of the PSR file

    Returns
    -------
    cpi_header, header : dict
        Dictionary containing the PSR header data and the CPI headers data

    """
    # load PSR library
    psr_lib = get_library()

    c_float_p = ctypes.POINTER(ctypes.c_float)
    c_long_p = ctypes.POINTER(ctypes.c_long)

    header = read_psr_header(filename)
    nitems = header['item.count']

    azi_start = np.empty(nitems, dtype=np.float32)
    azi_stop = np.empty(nitems, dtype=np.float32)
    ele_start = np.empty(nitems, dtype=np.float32)
    ele_stop = np.empty(nitems, dtype=np.float32)
    npulses = np.empty(nitems, dtype=np.int32)
    prfs = np.empty(nitems, dtype=np.float32)
    ngates = np.empty(nitems, dtype=np.int32)
    tx_pwr = np.empty(nitems, dtype=np.float32)

    c_filename = ctypes.c_char_p(filename.encode('utf-8'))

    try:
        psr_lib.psr_getValueArrays(
            c_filename, azi_start.ctypes.data_as(c_float_p),
            azi_stop.ctypes.data_as(c_float_p),
            ele_start.ctypes.data_as(c_float_p),
            ele_stop.ctypes.data_as(c_float_p),
            npulses.ctypes.data_as(c_long_p),
            prfs.ctypes.data_as(c_float_p),
            ngates.ctypes.data_as(c_long_p),
            tx_pwr.ctypes.data_as(c_float_p))
    except EnvironmentError as ee:
        warn(str(ee))
        warn('Unable to read file '+filename)
        return None, None

    noise = None
    if 'noise' in header:
        noise = np.ma.array(header['noise'])*npulses

    cpi_header = {
        'azi_start': azi_start,
        'azi_stop': azi_stop,
        'ele_start': ele_start,
        'ele_stop': ele_stop,
        'npulses': npulses,
        'prfs': prfs,
        'ngates': ngates,
        'tx_pwr': np.ma.array(tx_pwr, dtype=np.float),
        'noise': np.ma.array(noise)
    }

    return cpi_header, header


def read_psr_spectra(filename):
    """
    Reads the complex spectral data contained in a PSR file

    Parameters
    ----------
    filename : str
        Name of the PSR file

    Returns
    -------
    spectra : 3D complex ndArray
       The complex spectra

    """
    # load PSR library
    psr_lib = get_library()

    c_filename = ctypes.c_char_p(filename.encode('utf-8'))
    c_complex_p = np.ctypeslib.ndpointer(
        dtype=np.complex64, ndim=1, flags='C')

    cpi_header, _ = read_psr_cpi_header(filename)
    if cpi_header is None:
        return None

    nitems = cpi_header['ngates'].size
    ngates = np.max(cpi_header['ngates'])
    npulses_max = np.max(cpi_header['npulses'])

    spectra = np.ma.masked_all(
        (nitems, ngates, npulses_max), dtype=np.complex128)
    for item in range(nitems):
        npulses = cpi_header['npulses'][item]
        for gate in range(cpi_header['ngates'][item]):
            spectrum = np.empty(npulses, dtype=np.complex64)
            try:
                psr_lib.psr_getPowerSpectrum(
                    c_filename, ctypes.c_int(item), ctypes.c_int(gate),
                    spectrum.ctypes.data_as(c_complex_p))
                spectra[item, gate, 0:npulses] = spectrum
            except EnvironmentError as ee:
                warn(str(ee))
                warn('Unable to get CPI element '+str(item) +
                     ' at range gate '+str(gate)+' from file '+filename)

    return spectra


def get_item_numbers(radar, azi_start, azi_stop, ele_start, ele_stop,
                     prf_array, prfs, cpi='low_prf', ang_tol=0.5):
    """
    Gets the item numbers to be used and eventually modify the radar object
    to accomodate more angles


    Parameters
    ----------
    radar: radar object
        the reference radar
    azi_start, azi_stop, ele_start, ele_stop : float array
        The start and stop angles of the CPI elements
    prf_array: float array
        The PRF of each CPI element
    prfs : float array
        The unique PRFs contained in the PSR file
    cpi : str
        The CPI to use. Can be 'low_prf', 'intermediate_prf', 'high_prf',
        'mean', 'all'. If 'mean' the mean within the angle step is taken.
        If 'all' the data is not filtered by PRF
    ang_tol : float
        Angle tolerance

    Returns
    -------
    items : int array
        the item number selected

    """
    if radar.scan_type == 'ppi':
        cpi_ang_center = azi_start + (azi_stop-azi_start) / 2.
        cpi_fixed_angle = ele_start + (ele_stop-ele_start) / 2.

        ref_ang = radar.azimuth['data']
        fixed_ang = radar.elevation['data']

    elif radar.scan_type == 'rhi' or radar.scan_type == 'other':
        cpi_ang_center = ele_start + (ele_stop-ele_start) / 2.
        cpi_fixed_angle = azi_start + (azi_stop-azi_start) / 2.

        ref_ang = radar.elevation['data']
        fixed_ang = radar.azimuth['data']

    if prfs.size > 1 and cpi == 'all':
        items = np.array([], dtype=int)
        sweep_start = radar.sweep_start_ray_index['data']
        sweep_end = radar.sweep_end_ray_index['data']
        items_per_sweep = np.array([], dtype=int)

        items_aux = np.arange(cpi_ang_center.size)
        for i, (s_start, s_end) in enumerate(zip(sweep_start, sweep_end)):
            # only angles within fixed angle tolerance
            fixed = radar.fixed_angle['data'][i]
            ind = np.where(np.abs(cpi_fixed_angle-fixed) < ang_tol)[0]

            cpi_ang_center_aux = cpi_ang_center[ind]
            items_aux2 = items_aux[ind]

            # get angles within radar limits
            ind = np.where(np.logical_and(
                cpi_ang_center_aux >= ref_ang[s_start]-ang_tol,
                cpi_ang_center_aux <= ref_ang[s_end]+ang_tol))[0]

            items = np.append(items, items_aux2[ind])
            items_per_sweep = np.append(items_per_sweep, items_aux2[ind].size)

        radar = change_rays(
            radar, cpi_ang_center[items], cpi_fixed_angle[items],
            items_per_sweep)
        return items, radar

    items_aux = np.arange(cpi_ang_center.size)
    if prfs.size > 1:
        if cpi == 'low_prf':
            cpi_ang_center = cpi_ang_center[prf_array == prfs[0]]
            cpi_fixed_angle = cpi_fixed_angle[prf_array == prfs[0]]
            items_aux = items_aux[prf_array == prfs[0]]
        elif cpi == 'intermediate_prf':
            if prfs.size == 3:
                cpi_ang_center = cpi_ang_center[prf_array == prfs[1]]
                cpi_fixed_angle = cpi_fixed_angle[prf_array == prfs[1]]
                items_aux = items_aux[prf_array == prfs[1]]
            else:
                warn('Less than 3 different prfs. ' +
                     'Low prf data will be used')
                cpi_ang_center = cpi_ang_center[prf_array == prfs[0]]
                cpi_fixed_angle = cpi_fixed_angle[prf_array == prfs[0]]
                items_aux = items_aux[prf_array == prfs[0]]
        elif cpi == 'high_prf':
            cpi_ang_center = cpi_ang_center[prf_array == prfs[-1]]
            cpi_fixed_angle = cpi_fixed_angle[prf_array == prfs[-1]]
            items_aux = items_aux[prf_array == prfs[-1]]
        elif cpi == 'mean':
            cpi_ang_center_aux = []
            cpi_fixed_angle_aux = []
            items_aux2 = []
            for prf in prfs:
                cpi_ang_center_aux.append(cpi_ang_center[prf_array == prf])
                cpi_fixed_angle_aux.append(cpi_fixed_angle[prf_array == prf])
                items_aux2.append(items_aux[prf_array == prf])

            cpi_ang_center = cpi_ang_center_aux
            cpi_fixed_angle = cpi_fixed_angle_aux
            items_aux = items_aux2

    # get items to extract
    if prfs.size > 1 and cpi == 'mean':
        # fixed pointing scans
        if radar.scan_type == 'other':
            items = []
            for j, prf in enumerate(prfs):
                cpi_fixed_angle_aux = cpi_fixed_angle[j]
                cpi_ang_center_aux = cpi_ang_center[j]
                items_aux2 = items_aux[j]

                # only angles within fixed angle tolerance
                ind = np.where(
                    np.abs(cpi_fixed_angle_aux-fixed_ang[0]) < ang_tol)[0]
                if ind.size < radar.nrays:
                    return np.array([], dtype=int), radar
                items_aux4 = items_aux2[ind[:radar.nrays]]
                items.append(items_aux4)
            return items, radar

        for j, prf in enumerate(prfs):
            cpi_fixed_angle_aux = cpi_fixed_angle[j]
            cpi_ang_center_aux = cpi_ang_center[j]
            items_aux2 = items_aux[j]
            items_aux4 = np.array([], dtype=int)
            for i, ang in enumerate(ref_ang):
                # only angles within fixed angle tolerance
                fixed = fixed_ang[i]
                ind = np.where(np.abs(cpi_fixed_angle_aux-fixed) < ang_tol)
                cpi_ang_center_aux2 = cpi_ang_center_aux[ind]
                items_aux3 = items_aux2[ind]

                # get angle closest to reference angle
                ind = np.argmin(np.abs(cpi_ang_center_aux2-ang))
                items_aux4 = np.append(items_aux4, items_aux3[ind])
            items.append(items_aux4)

        return items, radar

    # fixed pointing scans
    if radar.scan_type == 'other':
        # only angles within fixed angle tolerance
        ind = np.where(np.abs(cpi_fixed_angle-fixed_ang[0]) < ang_tol)[0]
        if ind.size < radar.nrays:
            return np.array([], dtype=int), radar

        return items_aux[ind[:radar.nrays]], radar

    items = np.array([], dtype=int)
    for i, ang in enumerate(ref_ang):
        # only angles within fixed angle tolerance
        fixed = fixed_ang[i]
        ind = np.where(np.abs(cpi_fixed_angle-fixed) < ang_tol)[0]

        if ind.size == 0:
            continue

        cpi_ang_center_aux = cpi_ang_center[ind]
        items_aux2 = items_aux[ind]

        # get angle closest to reference angle
        ind = np.argmin(np.abs(cpi_ang_center_aux-ang))
        items = np.append(items, items_aux2[ind])

    return items, radar


def get_field(radar, cpi_header, header, items, nprfs, field_name,
              undo_txcorr=True, cpi='low_prf'):
    """
    Gets the field corresponding to the reference radar

    Parameters
    ----------
    radar: radar object
        the reference radar
    cpi_header, header : dict
        dictionaries containing the PSR file header and CPI headers data
    items : int array
        array containing the items to select
    nprfs: float
        The number of different prfs in the file
    field_name : str
        The name of the field to filter
    undo_txcorr : bool
        If True and field is a noise field the correction of the received
        signal by the transmitted power is undone
    cpi : str
        The CPI to use. Can be 'low_prf', 'intermediate_prf', 'high_prf',
        'all'. If 'all' the mean within the angle step is taken

    Returns
    -------
    field_data : 2D float array
        The PSR data in the format of the reference radar fields

    """
    if field_name in ('noisedBZ_hh', 'noisedBZ_vv', 'noisedBm_hh',
                      'noisedBm_vv', 'noisedBADU_hh', 'noisedBADU_vv'):
        field = cpi_header['noise']
        if undo_txcorr:
            field[cpi_header['tx_pwr'] > 0.] *= (
                cpi_header['tx_pwr'][cpi_header['tx_pwr'] > 0.] /
                header['states.gdrxmaxpowerkwpw'][header['states.spbpwidth']])

    elif field_name in ('transmitted_signal_power_h',
                        'transmitted_signal_power_v'):
        field = cpi_header['tx_pwr']

    if nprfs > 1 and cpi == 'mean':
        field_filt = np.ma.masked_all((radar.nrays, nprfs))
        npulses_filt = np.ma.masked_all((radar.nrays, nprfs), dtype=int)
        npulses = cpi_header['npulses']
        for i in range(nprfs):
            items_aux = items[i]
            for j, item in enumerate(items_aux):
                field_filt[j, i] = field[item]*npulses[item]
                npulses_filt[j, i] = npulses[item]
        field_filt = np.transpose(np.atleast_2d(
            np.ma.sum(field_filt, axis=-1)/np.ma.sum(npulses_filt, axis=-1)))
    else:
        field_filt = np.ma.masked_all((radar.nrays, 1))
        for i, item in enumerate(items):
            field_filt[i, 0] = field[item]

    field_data = ma_broadcast_to(field_filt, (radar.nrays, radar.ngates))

    return field_data


def get_spectral_noise(radar, cpi_header, header, items, undo_txcorr=True):
    """
    Gets the field corresponding to the reference radar

    Parameters
    ----------
    radar: radar object
        the reference radar
    cpi_header, header : dict
        dictionaries containing the PSR file header and CPI headers data
    items : int array
        array containing the items to select
    field_name : str
        The name of the field to filter
    undo_txcorr : bool
        If True and field is a noise field the correction of the received
        signal by the transmitted power is undone

    Returns
    -------
    field_data : 2D float array
        The PSR data in the format of the reference radar fields

    """
    field = cpi_header['noise']/cpi_header['npulses']
    if undo_txcorr:
        field[cpi_header['tx_pwr'] > 0.] *= (
            cpi_header['tx_pwr'][cpi_header['tx_pwr'] > 0.] /
            header['states.gdrxmaxpowerkwpw'][header['states.spbpwidth']])

    field_filt = np.ma.masked_all((radar.nrays, 1, 1))
    for i, item in enumerate(items):
        field_filt[i, 0, 0] = field[item]

    npulses_max = np.max(cpi_header['npulses'][items])

    field_data = ma_broadcast_to(field_filt, (radar.nrays, radar.ngates, npulses_max))

    return field_data


def get_spectra_field(radar, filenames, npulses, items_per_file, items,
                      ind_rng, fold=True, positive_away=True):
    """
    Gets the field corresponding to the reference radar


    Parameters
    ----------
    radar: radar object
        the reference radar
    filename : str
        name of the PSR file
    npulses : int array
        array containing the number of pulses for each item
    items_per_file : int array
        array containing the number of items in each PSR file
    items : int array
        array containing the items to select
    ind_rng : int array
        array containing the indices to the range gates to select
    fold : Bool
        If True the spectra is folded
    positive_away : Bool
        If True positive Doppler velocities are way from the radar

    Returns
    -------
    spectra : 3D complex float array
        The complex spectra field

    """
    # load PSR library
    psr_lib = get_library()

    c_complex_p = np.ctypeslib.ndpointer(
        dtype=np.complex64, ndim=1, flags='C')

    npulses_max = np.max(npulses[items])

    spectra = np.ma.masked_all(
        (radar.nrays, radar.ngates, npulses_max), dtype=np.complex64)

    accu_items = np.cumsum(items_per_file)
    for ray, item in enumerate(items):
        # get file to read and item number within file
        ind = np.where(accu_items >= item)[0]
        if ind.size == 0 or ind[0] == 0:
            item_aux = item
            ind = 0
        else:
            ind = ind[0]
            item_aux = item - accu_items[ind-1]

        c_filename = ctypes.c_char_p(filenames[ind].encode('utf-8'))
        for rng, gate in enumerate(ind_rng):
            npulses_item = npulses[item]
            spectrum = np.empty(npulses_item, dtype=np.complex64)
            try:
                psr_lib.psr_getPowerSpectrum(
                    c_filename, ctypes.c_int(item_aux), ctypes.c_int(gate),
                    spectrum.ctypes.data_as(c_complex_p))

                if fold:
                    nfold = int(np.ceil(npulses_item/2.))
                    spectrum = np.append(spectrum[nfold:], spectrum[0:nfold])
                if positive_away:
                    spectrum = spectrum[::-1]

                spectra[ray, rng, 0:npulses_item] = spectrum
            except EnvironmentError as ee:
                warn(str(ee))
                warn('Unable to get CPI element '+str(item) +
                     ' at range gate '+str(gate)+' from file '+filenames[ind])

    spectra = np.ma.masked_equal(spectra, 0.)

    return spectra


def get_Doppler_info(prfs, npulses, wavelength, fold=True):
    """
    Gets the Doppler information


    Parameters
    ----------
    prfs: float array
        the PRF at each ray
    npulses : float array
        the number of pulses per ray
    wavelength : float
        the radar wavelength [m]
    fold : Bool
        If True the spectra is folded

    Returns
    -------
    Doppler_velocity, Doppler_frequency : 2D float array
        The Doppler velocity and Doppler frequency bins for each ray

    """
    nrays = npulses.size
    npulses_max = np.max(npulses)
    freq_res = prfs/npulses
    vel_res = freq_res*wavelength/2.

    Doppler_frequency = np.ma.masked_all((nrays, npulses_max))
    Doppler_velocity = np.ma.masked_all((nrays, npulses_max))
    for ray in range(nrays):
        pulses_ray = np.arange(npulses[ray])
        npulses_ray = pulses_ray.size
        if fold:
            nfold = int(np.ceil(npulses_ray/2.))
            pulses_ray = np.append(
                pulses_ray[nfold:]-npulses_ray, pulses_ray[0:nfold])

        Doppler_frequency[ray, 0:npulses_ray] = pulses_ray*freq_res[ray]
        Doppler_velocity[ray, 0:npulses_ray] = pulses_ray*vel_res[ray]

    return Doppler_velocity, Doppler_frequency


def get_noise_field(radar, field_data, header, field_name):
    """
    Puts the noise field in the desired units

    Parameters
    ----------
    radar: radar object
        the reference radar
    field_data : 2D float array
        The PSR data in the format of the reference radar fields
    header : dict
        Dictionary containing the PSR file metadata
    field_name : str
        The name of the field

    Returns
    -------
    field_data : 2D float array
        The PSR data in the format of the reference radar fields

    """
    field_data = 10.*np.ma.log10(field_data)

    if field_name in ('noisedBADU_hh', 'noisedBADU_vv'):
        return field_data

    pw_ind = header['states.spbpwidth']
    if field_name in ('noisedBZ_hh', 'noisedBm_hh'):
        dBadu2dBm = header['states.spbdbmtologoffset'][pw_ind]
    else:
        dBadu2dBm = header['states.spbdpvdbmtologoffset'][pw_ind]

    field_data = field_data+dBadu2dBm

    if field_name in ('noisedBm_hh', 'noisedBm_vv'):
        return field_data

    if field_name == 'noisedBZ_hh':
        radconst = header['states.rspdphradconst'][pw_ind]
    else:
        radconst = header['states.rspdpvradconst'][pw_ind]
    mfloss = header['states.gdrxmfloss'][pw_ind]
    pathatt = header['states.rspathatt']

    rangeKm = ma_broadcast_to(
        np.atleast_2d(radar.range['data']/1000.), (radar.nrays, radar.ngates))

    field_data += radconst+mfloss+pathatt*rangeKm+20.*np.log10(rangeKm)

    return field_data


def convert_data(values):
    """
    Converts an string of values into the corresponding format

    Parameters
    ----------
    values: str
        string containg the values to convert

    Returns
    -------
    values : int, float, str or 1D array of int, float or str
        The converted values

    """
    if ' ' in values[1:-1]:
        values = values.split()
        try:
            return np.array(values, dtype=int)
        except ValueError:
            try:
                return np.array(values, dtype=float)
            except ValueError:
                return np.array(values, dtype=str)

    try:
        return int(values)
    except ValueError:
        try:
            return float(values)
        except ValueError:
            return values


def get_library():
    """
        return the link to C-shared library

        Returns
        -------
        psr_lib : link
            loaded PSR C-library

    """
    try:
        library_path = get_library_path()
    except MissingOptionalDependency:
        raise MissingOptionalDependency(" PSR library path NOT defined")

    try:
        psr_lib = ctypes.cdll.LoadLibrary(library_path+'/'+'libDX50.so')
    except OSError as ee:
        warn(str(ee))
        raise MissingOptionalDependency('Unable to load PSR library')

    return psr_lib


def get_library_path():
    """
    find valid library path

    Returns
    -------
    psr_lib_path : str
        library path

    """
    # Check if path is correct
    library_paths = [
        os.environ.get('PSRLIB_PATH'),
        os.path.expanduser('~')+'/pyrad/src/libDX50/lib/']
    library_path = ''
    for p in library_paths:
        if p is not None:
            if os.path.isdir(p):
                library_path = p
                break

    if not library_path:
        raise MissingOptionalDependency(" PSR library path NOT defined")

    return library_path


def change_rays(radar, moving_angle, fixed_angle, rays_per_sweep):
    """
    Modify the radar object to accomodate new rays

    Parameters
    ----------
    radar : radar object
        the radar to modify
    moving_angle : float array
        The moving angles
    fixed_angle : float array
        The fixed angles
    rays_per_sweep : array of ints
        The number of rays per sweep

    Returns
    -------
    new_radar : radar object
        The modified radar

    """
    new_radar = deepcopy(radar)
    new_radar.nrays = moving_angle.size
    if radar.scan_type == 'ppi':
        new_radar.azimuth['data'] = moving_angle
        new_radar.elevation['data'] = fixed_angle
    elif radar.scan_type == 'rhi':
        new_radar.azimuth['data'] = fixed_angle
        new_radar.elevation['data'] = moving_angle

    ray_factor = int(new_radar.nrays/radar.nrays)

    # change time
    time_res = np.append(
        (radar.time['data'][1:]-radar.time['data'][:-1])/ray_factor,
        (radar.time['data'][-1]-radar.time['data'][-2])/ray_factor)
    time_sum = np.cumsum(
        np.reshape(
            np.repeat(time_res, ray_factor), (radar.nrays, ray_factor)),
        axis=1).flatten()
    new_radar.time['data'] = (
        np.repeat(radar.time['data'], ray_factor)+ time_sum)

    new_radar.sweep_start_ray_index['data'] = np.append(
        0, np.cumsum(rays_per_sweep[:-1]))
    new_radar.sweep_end_ray_index['data'] = np.cumsum(rays_per_sweep-1)
    new_radar.init_rays_per_sweep()
    if new_radar.ray_angle_res is not None:
        new_radar.ray_angle_res['data'] /= ray_factor

    new_radar.init_gate_x_y_z()
    new_radar.init_gate_longitude_latitude()
    new_radar.init_gate_altitude()

    return new_radar



# /usr/bin/env python
# dummify (replace all data with a single value) a NEXRAD Level II file.

import numpy as np
import pyart.io.nexrad_level2 as nexrad
import struct

NEXRAD_FILE = 'KATX20130717_195021_V06'
OUTPUT_FILE = 'KATX20130717_195021_V06_DUMMY'

# sizes of structures
RECORD_SIZE = 2432
MSG_HEADER_SIZE = 16
DATA_BLOCK_SIZE = 28

fin = open(NEXRAD_FILE, 'rb')
out = open(OUTPUT_FILE, 'wb')

# read in the first 134 records as write as is
out.write(fin.read(24 + 12 + RECORD_SIZE * 134))

# read the rest of the file, and create a array of int8
buf = fin.read()
buf_length = len(buf)
buf_data = np.frombuffer(buf, dtype='i1')
fin.close()

# replace all radial data with a single value so it compresses well
pos = 0
while pos < buf_length:

    # retrieve the record (msg31)
    new_pos, record = nexrad._get_record_from_buf(buf, pos)
    if record['header']['type'] != 31:
        pos += RECORD_SIZE
        continue

    # dummy out all moments
    for bp_number in [4, 5, 6, 7, 8, 9]:
        bp_str = 'block_pointer_' + str(bp_number)
        block_pointer = record['msg_header'][bp_str]
        if block_pointer != 0:
            bpos = pos + block_pointer + MSG_HEADER_SIZE
            d = nexrad._unpack_from_buf(buf, bpos, nexrad.GENERIC_DATA_BLOCK)
            points = d['ngates']
            if d['data_name'] == 'PHI':
                points *= 2
            start = pos + MSG_HEADER_SIZE + DATA_BLOCK_SIZE + block_pointer
            buf_data[start:start+points] = 2

    pos = new_pos

# write out the modified data
out.write(buf_data.tostring())
out.close()



"""
pyart.aux_io.noxp_iphex_nc
==========================

Routines for reading IPHEx NOXP files.

.. autosummary::
    :toctree: generated/

    read_noxp_iphex_nc
    _ncvar_to_dict

"""

import datetime

import numpy as np
import netCDF4

from ..config import FileMetadata
from ..io.common import make_time_unit_str  # , _test_arguments
from ..core.radar import Radar

# Only providing updated names for the most common fields.
# Many more fields in the files, and these are read in but not renamed.
NOXP_FIELD_NAMES = {
    # uncorrected reflectivity, horizontal
    'ZH': 'reflectivity',
    # uncorrected differential reflectivity
    'ZDR': 'differential_reflectivity',
    'RHOHV': 'cross_correlation_ratio',
    'KDP': 'differential_phase',
    'KDPM': 'specific_differential_phase',
    'VR': 'velocity',
    'SVR': 'spectrum_width',
}


def read_noxp_iphex_nc(filename, field_names=None, additional_metadata=None,
                       file_field_names=False, exclude_fields=None, **kwargs):
    """
    Read a NOXP IPHEX netCDF file.

    Parameters
    ----------
    filename : str
        Name of the netCDF file to read.
    field_names : dict, optional
        Dictionary mapping netCDF field names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        Py-ART configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata from during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    file_field_names : bool, optional
        True to use the netCDF data type names for the field names. If this
        case the field_names parameter is ignored. The field dictionary will
        likely only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.

    Returns
    -------
    radar : Radar
        Radar object containing data from netCDF file.

    """
    # TODO before moving to pyart.io
    # * unit test
    # * add default field mapping, etc to default config
    # * auto-detect file type with pyart.io.read function
    # * instrument parameters
    # * add additional checks for HOW attributes
    # * support for other objects (SCAN, XSEC)

    # test for non empty kwargs
    # Getting import error on this function, skipping for now.
    # Only issues a warning anyway.
#    _test_arguments(kwargs)

    # create metadata retrieval object
    if field_names is None:
        field_names = NOXP_FIELD_NAMES
    filemetadata = FileMetadata('cfradial', field_names, additional_metadata,
                                file_field_names, exclude_fields,
                                include_fields)

    # read the data
    ncobj = netCDF4.Dataset(filename)
    ncvars = ncobj.variables

    # One sweep per file
    nsweeps = 1

    # latitude, longitude and altitude
    latitude = filemetadata('latitude')
    longitude = filemetadata('longitude')
    altitude = filemetadata('altitude')
    latitude['data'] = np.array(ncvars['Latitude'])
    longitude['data'] = np.array(ncvars['Longitude'])
    altitude['data'] = np.array(ncvars['Altitude'])

    # metadata
    metadata = filemetadata('metadata')
    metadata['source'] = 'NOAA NSSL'
    metadata['original_container'] = 'noxp_iphex_nc'
    metadata['system'] = ncobj.Radar

    # sweep_start_ray_index, sweep_end_ray_index
    sweep_start_ray_index = filemetadata('sweep_start_ray_index')
    sweep_end_ray_index = filemetadata('sweep_end_ray_index')
    rays_per_sweep = np.shape(ncvars['AZ'][:])
    ssri = np.cumsum(np.append([0], rays_per_sweep[:-1])).astype('int32')
    seri = np.cumsum(rays_per_sweep).astype('int32') - 1
    sweep_start_ray_index['data'] = ssri
    sweep_end_ray_index['data'] = seri

    # sweep_number
    sweep_number = filemetadata('sweep_number')
    sweep_number['data'] = np.arange(nsweeps, dtype='int32')

    # sweep_mode
    sweep_mode = filemetadata('sweep_mode')
    scan_type = getattr(ncobj, 'Scan type').lower()
    if scan_type in ['ppi', 'rhi']:
        sweep_mode['data'] = np.array(nsweeps * ['manual_' + scan_type])
    else:  # Guessing that if not RHI or PPI, then a pointing scan
        sweep_mode['data'] = np.array(nsweeps * ['pointing'])

    # fixed_angle
    fixed_angle = filemetadata('fixed_angle')
    if scan_type == 'rhi':
        sweep_el = ncvars['AZ'][0]
    else:
        sweep_el = ncvars['EL'][0]
    fixed_angle['data'] = np.array([sweep_el], dtype='float32')

    # elevation
    elevation = filemetadata('elevation')
    elevation['data'] = np.array(ncvars['EL'])

    # range
    _range = filemetadata('range')
    rng = 1000.0 * ncvars['Range'][:].T
    _range['data'] = rng[0]
    _range['meters_to_center_of_first_gate'] = np.mean(rng[0][0:2])
    # Gate spacing mostly but not completely constant
    _range['meters_between_gates'] = np.median(np.diff(_range['data']))
    _range['spacing_is_constant'] = 0

    # azimuth
    azimuth = filemetadata('azimuth')
    azimuth['data'] = np.array(ncvars['AZ'])

    # time
    # Basing everything around 6/3/2014 because better than 1/0/0000
    # for datetime module.
    _time = filemetadata('time')
    frac_since_basetime = ncvars['time'][0] - 735762.0
    start_time = \
        datetime.datetime(2014, 6, 3) + datetime.timedelta(frac_since_basetime)
    _time['units'] = make_time_unit_str(start_time)
    _time['data'] = 3600.0 * 24.0 * \
        (ncvars['time'][:]-ncvars['time'][0]).astype('float32')

    # fields
    # nearly all variables w/ dimensions of 'Gate', 'Time' are fields
    keys = [k for k, v in ncvars.items()
            if v.dimensions == ('Gate', 'Time') and
            k not in ['Range', 'Distance', 'Height']]
    fields = {}
    for key in keys:
        field_name = filemetadata.get_field_name(key)
        if field_name is None:
            if exclude_fields is not None and key in exclude_fields:
                continue
            if include_fields is not None and not key in include_fields:
                continue
            field_name = key
        fields[field_name] = _ncvar_to_dict(ncvars[key])

    # instrument_parameters
    instrument_parameters = {}
    prt = filemetadata('prt')
    prt['data'] = 1.0 / float(ncobj.PRF[0:5])
    instrument_parameters['prt'] = prt
    frequency = filemetadata('frequency')
    frequency['frequency'] = float(ncobj.Frequency[0:4]) * 10e9
    instrument_parameters['frequency'] = frequency
    hits = filemetadata('n_samples')
    hits['data'] = ncobj.Hits
    instrument_parameters['hits'] = hits

    # Go get the radar object!
    return Radar(
        _time, _range, fields, metadata, scan_type,
        latitude, longitude, altitude,
        sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth, elevation,
        instrument_parameters=instrument_parameters)


def _ncvar_to_dict(ncvar):
    """ Convert a NetCDF Dataset variable to a dictionary. """
    # copy all attribute except for scaling parameters
    d = dict((k, getattr(ncvar, k)) for k in ncvar.ncattrs()
             if k not in ['scale_factor', 'add_offset'])
    d['data'] = ncvar[:].T  # Data originally stored as (gate, azimuth)
    if np.isscalar(d['data']):
        # netCDF4 1.1.0+ returns a scalar for 0-dim array, we always want
        # 1-dim+ arrays with a valid shape.
        d['data'] = np.array(d['data'])
        d['data'].shape = (1, )
    return d
